# -*- coding: utf-8 -*-
# Licensed to the Apache Software Foundation (ASF) under one or more§
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import socket
import sys
import ssl

from requests.exceptions import ConnectTimeout

from mock import Mock, patch

import requests_mock

import libcloud.common.base

from libcloud.test import unittest
from libcloud.common.base import Connection, CertificateConnection
from libcloud.common.exceptions import RateLimitReachedError
from libcloud.http import LibcloudBaseConnection
from libcloud.http import LibcloudConnection
from libcloud.http import SignedHTTPSAdapter
from libcloud.utils.retry import Retry
from libcloud.utils.retry import RetryForeverOnRateLimitError
from libcloud.utils.retry import RETRY_EXCEPTIONS
from libcloud.utils.py3 import assertRaisesRegex


class BaseConnectionClassTestCase(unittest.TestCase):

    def setUp(self):
        self.orig_proxy = os.environ.pop('http_proxy', None)

    def tearDown(self):
        if self.orig_proxy:
            os.environ['http_proxy'] = self.orig_proxy
        elif 'http_proxy' in os.environ:
            del os.environ['http_proxy']

        libcloud.common.base.ALLOW_PATH_DOUBLE_SLASHES = False

    @classmethod
    def tearDownClass(cls):
        if 'http_proxy' in os.environ:
            del os.environ['http_proxy']

        libcloud.common.base.ALLOW_PATH_DOUBLE_SLASHES = False

    def test_parse_proxy_url(self):
        conn = LibcloudBaseConnection()

        proxy_url = 'http://127.0.0.1:3128'
        result = conn._parse_proxy_url(proxy_url=proxy_url)
        self.assertEqual(result[0], 'http')
        self.assertEqual(result[1], '127.0.0.1')
        self.assertEqual(result[2], 3128)
        self.assertIsNone(result[3])
        self.assertIsNone(result[4])

        proxy_url = 'https://127.0.0.2:3129'
        result = conn._parse_proxy_url(proxy_url=proxy_url)
        self.assertEqual(result[0], 'https')
        self.assertEqual(result[1], '127.0.0.2')
        self.assertEqual(result[2], 3129)
        self.assertIsNone(result[3])
        self.assertIsNone(result[4])

        proxy_url = 'http://user1:pass1@127.0.0.1:3128'
        result = conn._parse_proxy_url(proxy_url=proxy_url)
        self.assertEqual(result[0], 'http')
        self.assertEqual(result[1], '127.0.0.1')
        self.assertEqual(result[2], 3128)
        self.assertEqual(result[3], 'user1')
        self.assertEqual(result[4], 'pass1')

        proxy_url = 'https://user1:pass1@127.0.0.2:3129'
        result = conn._parse_proxy_url(proxy_url=proxy_url)
        self.assertEqual(result[0], 'https')
        self.assertEqual(result[1], '127.0.0.2')
        self.assertEqual(result[2], 3129)
        self.assertEqual(result[3], 'user1')
        self.assertEqual(result[4], 'pass1')

        proxy_url = 'http://127.0.0.1'
        expected_msg = 'proxy_url must be in the following format'
        assertRaisesRegex(self, ValueError, expected_msg,
                          conn._parse_proxy_url,
                          proxy_url=proxy_url)

        proxy_url = 'http://@127.0.0.1:3128'
        expected_msg = 'URL is in an invalid format'
        assertRaisesRegex(self, ValueError, expected_msg,
                          conn._parse_proxy_url,
                          proxy_url=proxy_url)

        proxy_url = 'http://user@127.0.0.1:3128'
        expected_msg = 'URL is in an invalid format'
        assertRaisesRegex(self, ValueError, expected_msg,
                          conn._parse_proxy_url,
                          proxy_url=proxy_url)

    def test_constructor(self):
        proxy_url = 'http://127.0.0.2:3128'
        os.environ['http_proxy'] = proxy_url
        conn = LibcloudConnection(host='localhost', port=80)
        self.assertEqual(conn.proxy_scheme, 'http')
        self.assertEqual(conn.proxy_host, '127.0.0.2')
        self.assertEqual(conn.proxy_port, 3128)
        self.assertEqual(conn.session.proxies, {
            'http': 'http://127.0.0.2:3128',
            'https': 'http://127.0.0.2:3128',
        })

        _ = os.environ.pop('http_proxy', None)
        conn = LibcloudConnection(host='localhost', port=80)
        self.assertIsNone(conn.proxy_scheme)
        self.assertIsNone(conn.proxy_host)
        self.assertIsNone(conn.proxy_port)

        proxy_url = 'http://127.0.0.3:3128'
        conn.set_http_proxy(proxy_url=proxy_url)
        self.assertEqual(conn.proxy_scheme, 'http')
        self.assertEqual(conn.proxy_host, '127.0.0.3')
        self.assertEqual(conn.proxy_port, 3128)
        self.assertEqual(conn.session.proxies, {
            'http': 'http://127.0.0.3:3128',
            'https': 'http://127.0.0.3:3128',
        })

        proxy_url = 'http://127.0.0.4:3128'
        conn = LibcloudConnection(host='localhost', port=80,
                                  proxy_url=proxy_url)
        self.assertEqual(conn.proxy_scheme, 'http')
        self.assertEqual(conn.proxy_host, '127.0.0.4')
        self.assertEqual(conn.proxy_port, 3128)
        self.assertEqual(conn.session.proxies, {
            'http': 'http://127.0.0.4:3128',
            'https': 'http://127.0.0.4:3128',
        })

        os.environ['http_proxy'] = proxy_url
        proxy_url = 'http://127.0.0.5:3128'
        conn = LibcloudConnection(host='localhost', port=80,
                                  proxy_url=proxy_url)
        self.assertEqual(conn.proxy_scheme, 'http')
        self.assertEqual(conn.proxy_host, '127.0.0.5')
        self.assertEqual(conn.proxy_port, 3128)
        self.assertEqual(conn.session.proxies, {
            'http': 'http://127.0.0.5:3128',
            'https': 'http://127.0.0.5:3128',
        })

        os.environ['http_proxy'] = proxy_url
        proxy_url = 'https://127.0.0.6:3129'
        conn = LibcloudConnection(host='localhost', port=80,
                                  proxy_url=proxy_url)
        self.assertEqual(conn.proxy_scheme, 'https')
        self.assertEqual(conn.proxy_host, '127.0.0.6')
        self.assertEqual(conn.proxy_port, 3129)
        self.assertEqual(conn.session.proxies, {
            'http': 'https://127.0.0.6:3129',
            'https': 'https://127.0.0.6:3129',
        })

    def test_connection_to_unusual_port(self):
        conn = LibcloudConnection(host='localhost', port=8080)
        self.assertIsNone(conn.proxy_scheme)
        self.assertIsNone(conn.proxy_host)
        self.assertIsNone(conn.proxy_port)
        self.assertEqual(conn.host, 'http://localhost:8080')

        conn = LibcloudConnection(host='localhost', port=80)
        self.assertEqual(conn.host, 'http://localhost')

    def test_connection_session_timeout(self):
        """
        Test that the connection timeout attribute is set correctly
        """
        conn = LibcloudConnection(host='localhost', port=8080)
        self.assertEqual(conn.session.timeout, 60)

        conn = LibcloudConnection(host='localhost', port=8080, timeout=10)
        self.assertEqual(conn.session.timeout, 10)

    def test_connection_timeout_raised(self):
        """
        Test that the connection times out
        """
        conn = LibcloudConnection(host='localhost', port=8080, timeout=0.1)
        # use a not-routable address to test that the connection timeouts
        host = "http://10.255.255.1"
        with self.assertRaises(ConnectTimeout):
            conn.request('GET', host)

    def test_connection_url_merging(self):
        """
        Test that the connection class will parse URLs correctly
        """
        conn = Connection(url='http://test.com/')
        conn.connect()
        self.assertEqual(conn.connection.host, 'http://test.com')
        with requests_mock.mock() as m:
            m.get('http://test.com/test', text='data')
            response = conn.request('/test')
        self.assertEqual(response.body, 'data')

    def test_morph_action_hook(self):
        conn = Connection(url="http://test.com")

        conn.request_path = ''
        self.assertEqual(conn.morph_action_hook('/test'), '/test')
        self.assertEqual(conn.morph_action_hook('test'), '/test')

        conn.request_path = '/v1'
        self.assertEqual(conn.morph_action_hook('/test'), '/v1/test')
        self.assertEqual(conn.morph_action_hook('test'), '/v1/test')

        conn.request_path = '/v1'
        self.assertEqual(conn.morph_action_hook('/test'), '/v1/test')
        self.assertEqual(conn.morph_action_hook('test'), '/v1/test')

        conn.request_path = 'v1'
        self.assertEqual(conn.morph_action_hook('/test'), '/v1/test')
        self.assertEqual(conn.morph_action_hook('test'), '/v1/test')

        conn.request_path = 'v1/'
        self.assertEqual(conn.morph_action_hook('/test'), '/v1/test')
        self.assertEqual(conn.morph_action_hook('test'), '/v1/test')

        conn.request_path = '/a'
        self.assertEqual(conn.morph_action_hook('//b/c.txt'), '/a/b/c.txt')

        conn.request_path = '/b'
        self.assertEqual(conn.morph_action_hook('/foo//'), '/b/foo/')

        libcloud.common.base.ALLOW_PATH_DOUBLE_SLASHES = True

        conn.request_path = '/'
        self.assertEqual(conn.morph_action_hook('/'), '//')

        conn.request_path = ''
        self.assertEqual(conn.morph_action_hook('/'), '/')

        conn.request_path = '/a'
        self.assertEqual(conn.morph_action_hook('//b/c.txt'), '/a//b/c.txt')

        conn.request_path = '/b'
        self.assertEqual(conn.morph_action_hook('/foo//'), '/b/foo//')

    def test_connect_with_prefix(self):
        """
        Test that a connection with a base path (e.g. /v1/) will
        add the base path to requests
        """
        conn = Connection(url='http://test.com/')
        conn.connect()
        conn.request_path = '/v1'
        self.assertEqual(conn.connection.host, 'http://test.com')
        with requests_mock.mock() as m:
            m.get('http://test.com/v1/test', text='data')
            response = conn.request('/test')
        self.assertEqual(response.body, 'data')

    def test_secure_connection_unusual_port(self):
        """
        Test that the connection class will default to secure (https) even
        when the port is an unusual (non 443, 80) number
        """
        conn = Connection(secure=True, host='localhost', port=8081)
        conn.connect()
        self.assertEqual(conn.connection.host, 'https://localhost:8081')

        conn2 = Connection(url='https://localhost:8081')
        conn2.connect()
        self.assertEqual(conn2.connection.host, 'https://localhost:8081')

    def test_secure_by_default(self):
        """
        Test that the connection class will default to secure (https)
        """
        conn = Connection(host='localhost', port=8081)
        conn.connect()
        self.assertEqual(conn.connection.host, 'https://localhost:8081')

    def test_implicit_port(self):
        """
        Test that the port is not included in the URL if the protocol implies
        the port, e.g. http implies 80
        """
        conn = Connection(secure=True, host='localhost', port=443)
        conn.connect()
        self.assertEqual(conn.connection.host, 'https://localhost')

        conn2 = Connection(secure=False, host='localhost', port=80)
        conn2.connect()
        self.assertEqual(conn2.connection.host, 'http://localhost')

    def test_insecure_connection_unusual_port(self):
        """
        Test that the connection will allow unusual ports and insecure
        schemes
        """
        conn = Connection(secure=False, host='localhost', port=8081)
        conn.connect()
        self.assertEqual(conn.connection.host, 'http://localhost:8081')

        conn2 = Connection(url='http://localhost:8081')
        conn2.connect()
        self.assertEqual(conn2.connection.host, 'http://localhost:8081')


class ConnectionClassTestCase(unittest.TestCase):
    def setUp(self):
        self.originalConnect = Connection.connect
        self.originalResponseCls = Connection.responseCls

        Connection.connect = Mock()
        Connection.responseCls = Mock()
        Connection.allow_insecure = True

    def tearDown(self):
        Connection.connect = self.originalConnect
        Connection.responseCls = Connection.responseCls
        Connection.allow_insecure = True

    def test_dont_allow_insecure(self):
        Connection.allow_insecure = True
        Connection(secure=False)

        Connection.allow_insecure = False

        expected_msg = (r'Non https connections are not allowed \(use '
                        r'secure=True\)')
        assertRaisesRegex(self, ValueError, expected_msg, Connection,
                          secure=False)

    def test_cache_busting(self):
        params1 = {'foo1': 'bar1', 'foo2': 'bar2'}
        params2 = [('foo1', 'bar1'), ('foo2', 'bar2')]

        con = Connection()
        con.connection = Mock()
        con.pre_connect_hook = Mock()
        con.pre_connect_hook.return_value = {}, {}
        con.cache_busting = False

        con.request(action='/path', params=params1)
        args, kwargs = con.pre_connect_hook.call_args
        self.assertFalse('cache-busting' in args[0])
        self.assertEqual(args[0], params1)

        con.request(action='/path', params=params2)
        args, kwargs = con.pre_connect_hook.call_args
        self.assertFalse('cache-busting' in args[0])
        self.assertEqual(args[0], params2)

        con.cache_busting = True

        con.request(action='/path', params=params1)
        args, kwargs = con.pre_connect_hook.call_args
        self.assertTrue('cache-busting' in args[0])

        con.request(action='/path', params=params2)
        args, kwargs = con.pre_connect_hook.call_args
        self.assertTrue('cache-busting' in args[0][len(params2)])

    def test_context_is_reset_after_request_has_finished(self):
        context = {'foo': 'bar'}

        def responseCls(connection, response):
            connection.called = True
            self.assertEqual(connection.context, context)

        con = Connection()
        con.called = False
        con.connection = Mock()
        con.responseCls = responseCls

        con.set_context(context)
        self.assertEqual(con.context, context)

        con.request('/')

        # Context should have been reset
        self.assertTrue(con.called)
        self.assertEqual(con.context, {})

        # Context should also be reset if a method inside request throws
        con = Connection(timeout=1, retry_delay=0.1)
        con.connection = Mock()

        con.set_context(context)
        self.assertEqual(con.context, context)
        con.connection.request = Mock(side_effect=ssl.SSLError())

        try:
            con.request('/')
        except ssl.SSLError:
            pass

        self.assertEqual(con.context, {})

        con.connection = Mock()
        con.set_context(context)
        self.assertEqual(con.context, context)

        con.responseCls = Mock(side_effect=ValueError())

        try:
            con.request('/')
        except ValueError:
            pass

        self.assertEqual(con.context, {})

    def _raise_socket_error(self):
        raise socket.gaierror('')

    @patch('libcloud.common.base.Connection.request')
    def test_retry_with_sleep(self, mock_connect):
        con = Connection()
        con.connection = Mock()

        mock_connect.side_effect = socket.gaierror('')
        retry_request = Retry(timeout=0.2, retry_delay=0.1,
                                backoff=1)
        self.assertRaises(socket.gaierror,
                          retry_request(con.request), action='/')

        self.assertGreater(mock_connect.call_count, 1,
                            'Retry logic failed')

    @patch('libcloud.common.base.Connection.request')
    def test_retry_with_timeout(self, mock_connect):
        con = Connection()
        con.connection = Mock()

        mock_connect.side_effect = socket.gaierror('')
        retry_request = Retry(timeout=0.2, retry_delay=0.1,
                                backoff=1)
        self.assertRaises(socket.gaierror,
                          retry_request(con.request), action='/')

        self.assertGreater(mock_connect.call_count, 1,
                            'Retry logic failed')

    @patch('libcloud.common.base.Connection.request')
    def test_retry_with_backoff(self, mock_connect):
        con = Connection()
        con.connection = Mock()

        mock_connect.side_effect = socket.gaierror('')
        retry_request = Retry(timeout=0.2, retry_delay=0.1,
                                backoff=1)
        self.assertRaises(socket.gaierror,
                          retry_request(con.request), action='/')
        self.assertGreater(mock_connect.call_count, 1,
                            'Retry logic failed')

    @patch('libcloud.common.base.Connection.request')
    def test_retry_rate_limit_error_timeout(self, mock_connect):
        con = Connection()
        con.connection = Mock()

        mock_connect.__name__ = 'mock_connect'
        headers = {'retry-after': 0.2}
        mock_connect.side_effect = RateLimitReachedError(headers=headers)
        retry_request = Retry(timeout=0.4, retry_delay=0.1,
                                backoff=1)
        self.assertRaises(RateLimitReachedError,
                          retry_request(con.request), action='/')
        self.assertEqual(mock_connect.call_count, 2,
                        'Retry logic failed')

    @patch('libcloud.common.base.Connection.request')
    def test_retry_rate_limit_error_forever_with_old_retry_class(self, mock_connect):
        con = Connection()
        con.connection = Mock()

        self.retry_counter = 0

        def mock_connect_side_effect(*args, **kwargs):
            self.retry_counter += 1

            if self.retry_counter < 4:
                headers = {'retry-after': 0.1}
                raise RateLimitReachedError(headers=headers)

            return 'success'

        mock_connect.__name__ = 'mock_connect'
        headers = {'retry-after': 0.2}
        mock_connect.side_effect = mock_connect_side_effect
        retry_request = RetryForeverOnRateLimitError(timeout=0.1, retry_delay=0.1,
                                backoff=1)
        retry_request(con.request)(action='/')

        # We have waited longer the timeout but continue to retry
        result = retry_request(con.request)(action='/')
        self.assertEqual(result, "success")

        self.assertEqual(mock_connect.call_count, 5,
                        'Retry logic failed')

    @patch('libcloud.common.base.Connection.request')
    def test_retry_should_not_retry_on_non_defined_exception(self, mock_connect):
        con = Connection()
        con.connection = Mock()
        connect_method = 'libcloud.common.base.Connection.request'

        self.retry_counter = 0

        mock_connect.__name__ = 'mock_connect'
        headers = {'retry-after': 0.2}
        mock_connect.side_effect = ValueError('should not retry this '
                                            'error')
        retry_request = Retry(timeout=5, retry_delay=0.1, backoff=1)

        self.assertRaisesRegex(ValueError, 'should not retry this error',
                               retry_request(con.request), action='/')
        self.assertEqual(mock_connect.call_count, 1, 'Retry logic failed')

    @patch('libcloud.common.base.Connection.request')
    def test_retry_rate_limit_error_success_on_second_attempt(self, mock_connect):
        con = Connection()
        con.connection = Mock()
        connect_method = 'libcloud.common.base.Connection.request'

        self.retry_counter = 0

        def mock_connect_side_effect(*args, **kwargs):
            self.retry_counter += 1

            if self.retry_counter < 2:
                headers = {'retry-after': 0.2}
                raise RateLimitReachedError(headers=headers)

            return 'success'

        mock_connect.__name__ = 'mock_connect'
        mock_connect.side_effect = mock_connect_side_effect
        retry_request = Retry(timeout=0.6, retry_delay=0.1,
                                backoff=1)
        result = retry_request(con.request)(action='/')
        self.assertEqual(result, "success")

        self.assertEqual(mock_connect.call_count, 2,
                        'Retry logic failed')

    @patch('libcloud.common.base.Connection.request')
    def test_retry_on_all_default_retry_exception_classes(self, mock_connect):
        con = Connection()
        con.connection = Mock()
        connect_method = 'libcloud.common.base.Connection.request'

        self.retry_counter = 0

        def mock_connect_side_effect(*args, **kwargs):
            self.retry_counter += 1

            if self.retry_counter < len(RETRY_EXCEPTIONS):
                raise RETRY_EXCEPTIONS[self.retry_counter]

            return 'success'

        mock_connect.__name__ = 'mock_connect'
        mock_connect.side_effect = mock_connect_side_effect
        retry_request = Retry(timeout=0.6, retry_delay=0.1,
                                backoff=1)
        result = retry_request(con.request)(action='/')
        self.assertEqual(result, "success")

        self.assertEqual(mock_connect.call_count, len(RETRY_EXCEPTIONS),
                        'Retry logic failed')


class CertificateConnectionClassTestCase(unittest.TestCase):
    def setUp(self):
        self.connection = CertificateConnection(cert_file='test.pem',
                                                url='https://test.com/test')
        self.connection.connect()

    def test_adapter_internals(self):
        adapter = self.connection.connection.session.adapters['https://']
        self.assertTrue(isinstance(adapter, SignedHTTPSAdapter))
        self.assertEqual(adapter.cert_file, 'test.pem')


if __name__ == '__main__':
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Vultr Driver
"""
import time
import json
import base64
from functools import update_wrapper
from typing import Optional, List, Dict, Union, Any

from libcloud.common.base import ConnectionKey, JsonResponse
from libcloud.common.types import InvalidCredsError
from libcloud.common.types import LibcloudError
from libcloud.common.types import ServiceUnavailableError
from libcloud.compute.base import Node, NodeImage, NodeSize, NodeLocation
from libcloud.compute.base import KeyPair, StorageVolume
from libcloud.compute.base import NodeDriver
from libcloud.compute.types import NodeState, StorageVolumeState
from libcloud.compute.types import Provider, VolumeSnapshotState
from libcloud.utils.iso8601 import parse_date
from libcloud.utils.py3 import httplib
from libcloud.utils.py3 import urlencode
from libcloud.utils.publickey import get_pubkey_openssh_fingerprint
from libcloud.common.vultr import DEFAULT_API_VERSION
from libcloud.common.vultr import VultrConnectionV2
from libcloud.common.vultr import VultrNetwork, VultrNodeSnapshot

# For matching region by id
VULTR_COMPUTE_INSTANCE_LOCATIONS = {
    "1": {
        "DCID": "1",
        "name": "New Jersey",
        "country": "US",
        "continent": "North America",
        "state": "NJ",
        "regioncode": "EWR"
    },
    "2": {
        "DCID": "2",
        "name": "Chicago",
        "country": "US",
        "continent": "North America",
        "state": "IL",
        "regioncode": "ORD"
    },
    "3": {
        "DCID": "3",
        "name": "Dallas",
        "country": "US",
        "continent": "North America",
        "state": "TX",
        "regioncode": "DFW"
    },
    "4": {
        "DCID": "4",
        "name": "Seattle",
        "country": "US",
        "continent": "North America",
        "state": "WA",
        "regioncode": "SEA"
    },
    "5": {
        "DCID": "5",
        "name": "Los Angeles",
        "country": "US",
        "continent": "North America",
        "state": "CA",
        "regioncode": "LAX"
    },
    "6": {
        "DCID": "6",
        "name": "Atlanta",
        "country": "US",
        "continent": "North America",
        "state": "GA",
        "regioncode": "ATL"
    },
    "7": {
        "DCID": "7",
        "name": "Amsterdam",
        "country": "NL",
        "continent": "Europe",
        "state": "",
        "regioncode": "AMS"
    },
    "8": {
        "DCID": "8",
        "name": "London",
        "country": "GB",
        "continent": "Europe",
        "state": "",
        "regioncode": "LHR"
    },
    "9": {
        "DCID": "9",
        "name": "Frankfurt",
        "country": "DE",
        "continent": "Europe",
        "state": "",
        "regioncode": "FRA"
    },
    "12": {
        "DCID": "12",
        "name": "Silicon Valley",
        "country": "US",
        "continent": "North America",
        "state": "CA",
        "regioncode": "SJC"
    },
    "19": {
        "DCID": "19",
        "name": "Sydney",
        "country": "AU",
        "continent": "Australia",
        "state": "",
        "regioncode": "SYD"
    },
    "22": {
        "DCID": "22",
        "name": "Toronto",
        "country": "CA",
        "continent": "North America",
        "state": "",
        "regioncode": "YTO"
    },
    "24": {
        "DCID": "24",
        "name": "Paris",
        "country": "FR",
        "continent": "Europe",
        "state": "",
        "regioncode": "CDG"
    },
    "25": {
        "DCID": "25",
        "name": "Tokyo",
        "country": "JP",
        "continent": "Asia",
        "state": "",
        "regioncode": "NRT"
    },
    "34": {
        "DCID": "34",
        "name": "Seoul",
        "country": "KR",
        "continent": "Asia",
        "state": "",
        "regioncode": "ICN"
    },
    "39": {
        "DCID": "39",
        "name": "Miami",
        "country": "US",
        "continent": "North America",
        "state": "FL",
        "regioncode": "MIA"
    },
    "40": {
        "DCID": "40",
        "name": "Singapore",
        "country": "SG",
        "continent": "Asia",
        "state": "",
        "regioncode": "SGP"
    }
}
# For matching image by id
VULTR_COMPUTE_INSTANCE_IMAGES = {
    "127": {
        "OSID": 127,
        "name": "CentOS 6 x64",
        "arch": "x64",
        "family": "centos",
        "windows": False
    },
    "147": {
        "OSID": 147,
        "name": "CentOS 6 i386",
        "arch": "i386",
        "family": "centos",
        "windows": False
    },
    "167": {
        "OSID": 167,
        "name": "CentOS 7 x64",
        "arch": "x64",
        "family": "centos",
        "windows": False
    },
    "381": {
        "OSID": 381,
        "name": "CentOS 7 SELinux x64",
        "arch": "x64",
        "family": "centos",
        "windows": False
    },
    "362": {
        "OSID": 362,
        "name": "CentOS 8 x64",
        "arch": "x64",
        "family": "centos",
        "windows": False
    },
    "401": {
        "OSID": 401,
        "name": "CentOS 8 Stream x64",
        "arch": "x64",
        "family": "centos",
        "windows": False
    },
    "215": {
        "OSID": 215,
        "name": "Ubuntu 16.04 x64",
        "arch": "x64",
        "family": "ubuntu",
        "windows": False
    },
    "216": {
        "OSID": 216,
        "name": "Ubuntu 16.04 i386",
        "arch": "i386",
        "family": "ubuntu",
        "windows": False
    },
    "270": {
        "OSID": 270,
        "name": "Ubuntu 18.04 x64",
        "arch": "x64",
        "family": "ubuntu",
        "windows": False
    },
    "387": {
        "OSID": 387,
        "name": "Ubuntu 20.04 x64",
        "arch": "x64",
        "family": "ubuntu",
        "windows": False
    },
    "194": {
        "OSID": 194,
        "name": "Debian 8 i386 (jessie)",
        "arch": "i386",
        "family": "debian",
        "windows": False
    },
    "244": {
        "OSID": 244,
        "name": "Debian 9 x64 (stretch)",
        "arch": "x64",
        "family": "debian",
        "windows": False
    },
    "352": {
        "OSID": 352,
        "name": "Debian 10 x64 (buster)",
        "arch": "x64",
        "family": "debian",
        "windows": False
    },
    "230": {
        "OSID": 230,
        "name": "FreeBSD 11 x64",
        "arch": "x64",
        "family": "freebsd",
        "windows": False
    },
    "327": {
        "OSID": 327,
        "name": "FreeBSD 12 x64",
        "arch": "x64",
        "family": "freebsd",
        "windows": False
    },
    "366": {
        "OSID": 366,
        "name": "OpenBSD 6.6 x64",
        "arch": "x64",
        "family": "openbsd",
        "windows": False
    },
    "394": {
        "OSID": 394,
        "name": "OpenBSD 6.7 x64",
        "arch": "x64",
        "family": "openbsd",
        "windows": False
    },
    "391": {
        "OSID": 391,
        "name": "Fedora CoreOS",
        "arch": "x64",
        "family": "fedora-coreos",
        "windows": False
    },
    "367": {
        "OSID": 367,
        "name": "Fedora 31 x64",
        "arch": "x64",
        "family": "fedora",
        "windows": False
    },
    "389": {
        "OSID": 389,
        "name": "Fedora 32 x64",
        "arch": "x64",
        "family": "fedora",
        "windows": False
    },
    "124": {
        "OSID": 124,
        "name": "Windows 2012 R2 x64",
        "arch": "x64",
        "family": "windows",
        "windows": False
    },
    "240": {
        "OSID": 240,
        "name": "Windows 2016 x64",
        "arch": "x64",
        "family": "windows",
        "windows": False
    },
    "159": {
        "OSID": 159,
        "name": "Custom",
        "arch": "x64",
        "family": "iso",
        "windows": False
    },
    "164": {
        "OSID": 164,
        "name": "Snapshot",
        "arch": "x64",
        "family": "snapshot",
        "windows": False
    },
    "180": {
        "OSID": 180,
        "name": "Backup",
        "arch": "x64",
        "family": "backup",
        "windows": False
    },
    "186": {
        "OSID": 186,
        "name": "Application",
        "arch": "x64",
        "family": "application",
        "windows": False
    }
}
VULTR_COMPUTE_INSTANCE_SIZES = {
    "201": {
        "VPSPLANID": "201",
        "name": "1024 MB RAM,25 GB SSD,1.00 TB BW",
        "vcpu_count": "1",
        "ram": "1024",
        "disk": "25",
        "bandwidth": "1.00",
        "bandwidth_gb": "1024",
        "price_per_month": "5.00",
        "plan_type": "SSD",
        "windows": False,
    },
    "202": {
        "VPSPLANID": "202",
        "name": "2048 MB RAM,55 GB SSD,2.00 TB BW",
        "vcpu_count": "1",
        "ram": "2048",
        "disk": "55",
        "bandwidth": "2.00",
        "bandwidth_gb": "2048",
        "price_per_month": "10.00",
        "plan_type": "SSD",
        "windows": False,
    },
    "203": {
        "VPSPLANID": "203",
        "name": "4096 MB RAM,80 GB SSD,3.00 TB BW",
        "vcpu_count": "2",
        "ram": "4096",
        "disk": "80",
        "bandwidth": "3.00",
        "bandwidth_gb": "3072",
        "price_per_month": "20.00",
        "plan_type": "SSD",
        "windows": False,
    },
    "204": {
        "VPSPLANID": "204",
        "name": "8192 MB RAM,160 GB SSD,4.00 TB BW",
        "vcpu_count": "4",
        "ram": "8192",
        "disk": "160",
        "bandwidth": "4.00",
        "bandwidth_gb": "4096",
        "price_per_month": "40.00",
        "plan_type": "SSD",
        "windows": False,
    },
    "205": {
        "VPSPLANID": "205",
        "name": "16384 MB RAM,320 GB SSD,5.00 TB BW",
        "vcpu_count": "6",
        "ram": "16384",
        "disk": "320",
        "bandwidth": "5.00",
        "bandwidth_gb": "5120",
        "price_per_month": "80.00",
        "plan_type": "SSD",
        "windows": False,
    },
    "206": {
        "VPSPLANID": "206",
        "name": "32768 MB RAM,640 GB SSD,6.00 TB BW",
        "vcpu_count": "8",
        "ram": "32768",
        "disk": "640",
        "bandwidth": "6.00",
        "bandwidth_gb": "6144",
        "price_per_month": "160.00",
        "plan_type": "SSD",
        "windows": False,
    },
    "207": {
        "VPSPLANID": "207",
        "name": "65536 MB RAM,1280 GB SSD,10.00 TB BW",
        "vcpu_count": "16",
        "ram": "65536",
        "disk": "1280",
        "bandwidth": "10.00",
        "bandwidth_gb": "10240",
        "price_per_month": "320.00",
        "plan_type": "SSD",
        "windows": False,
    },
    "208": {
        "VPSPLANID": "208",
        "name": "98304 MB RAM,1600 GB SSD,15.00 TB BW",
        "vcpu_count": "24",
        "ram": "98304",
        "disk": "1600",
        "bandwidth": "15.00",
        "bandwidth_gb": "15360",
        "price_per_month": "640.00",
        "plan_type": "SSD",
        "windows": False,
    },
    "115": {
        "VPSPLANID": "115",
        "name": "8192 MB RAM,110 GB SSD,10.00 TB BW",
        "vcpu_count": "2",
        "ram": "8192",
        "disk": "110",
        "bandwidth": "10.00",
        "bandwidth_gb": "10240",
        "price_per_month": "60.00",
        "plan_type": "DEDICATED",
        "windows": False,
    },
    "116": {
        "VPSPLANID": "116",
        "name": "16384 MB RAM,2x110 GB SSD,20.00 TB BW",
        "vcpu_count": "4",
        "ram": "16384",
        "disk": "110",
        "bandwidth": "20.00",
        "bandwidth_gb": "20480",
        "price_per_month": "120.00",
        "plan_type": "DEDICATED",
        "windows": False,
    },
    "117": {
        "VPSPLANID": "117",
        "name": "24576 MB RAM,3x110 GB SSD,30.00 TB BW",
        "vcpu_count": "6",
        "ram": "24576",
        "disk": "110",
        "bandwidth": "30.00",
        "bandwidth_gb": "30720",
        "price_per_month": "180.00",
        "plan_type": "DEDICATED",
        "windows": False,
    },
    "118": {
        "VPSPLANID": "118",
        "name": "32768 MB RAM,4x110 GB SSD,40.00 TB BW",
        "vcpu_count": "8",
        "ram": "32768",
        "disk": "110",
        "bandwidth": "40.00",
        "bandwidth_gb": "40960",
        "price_per_month": "240.00",
        "plan_type": "DEDICATED",
        "windows": False,
    },
    "400": {
        "VPSPLANID": "400",
        "name": "1024 MB RAM,32 GB SSD,1.00 TB BW",
        "vcpu_count": "1",
        "ram": "1024",
        "disk": "32",
        "bandwidth": "1.00",
        "bandwidth_gb": "1024",
        "price_per_month": "6.00",
        "plan_type": "HIGHFREQUENCY",
        "windows": False,
    },
    "401": {
        "VPSPLANID": "401",
        "name": "2048 MB RAM,64 GB SSD,2.00 TB BW",
        "vcpu_count": "1",
        "ram": "2048",
        "disk": "64",
        "bandwidth": "2.00",
        "bandwidth_gb": "2048",
        "price_per_month": "12.00",
        "plan_type": "HIGHFREQUENCY",
        "windows": False,
    },
    "402": {
        "VPSPLANID": "402",
        "name": "4096 MB RAM,128 GB SSD,3.00 TB BW",
        "vcpu_count": "2",
        "ram": "4096",
        "disk": "128",
        "bandwidth": "3.00",
        "bandwidth_gb": "3072",
        "price_per_month": "24.00",
        "plan_type": "HIGHFREQUENCY",
        "windows": False,
    },
    "403": {
        "VPSPLANID": "403",
        "name": "8192 MB RAM,256 GB SSD,4.00 TB BW",
        "vcpu_count": "3",
        "ram": "8192",
        "disk": "256",
        "bandwidth": "4.00",
        "bandwidth_gb": "4096",
        "price_per_month": "48.00",
        "plan_type": "HIGHFREQUENCY",
        "windows": False,
    },
    "404": {
        "VPSPLANID": "404",
        "name": "16384 MB RAM,384 GB SSD,5.00 TB BW",
        "vcpu_count": "4",
        "ram": "16384",
        "disk": "384",
        "bandwidth": "5.00",
        "bandwidth_gb": "5120",
        "price_per_month": "96.00",
        "plan_type": "HIGHFREQUENCY",
        "windows": False,
    },
    "405": {
        "VPSPLANID": "405",
        "name": "32768 MB RAM,512 GB SSD,6.00 TB BW",
        "vcpu_count": "8",
        "ram": "32768",
        "disk": "512",
        "bandwidth": "6.00",
        "bandwidth_gb": "6144",
        "price_per_month": "192.00",
        "plan_type": "HIGHFREQUENCY",
        "windows": False,
    },
    "406": {
        "VPSPLANID": "406",
        "name": "49152 MB RAM,768 GB SSD,8.00 TB BW",
        "vcpu_count": "12",
        "ram": "49152",
        "disk": "768",
        "bandwidth": "8.00",
        "bandwidth_gb": "8192",
        "price_per_month": "256.00",
        "plan_type": "HIGHFREQUENCY",
        "windows": False,
    }
}


class rate_limited:
    """
    Decorator for retrying Vultr calls that are rate-limited.

    :param int sleep: Seconds to sleep after being rate-limited.
    :param int retries: Number of retries.
    """

    def __init__(self, sleep=0.5, retries=1):
        self.sleep = sleep
        self.retries = retries

    def __call__(self, call):
        """
        Run ``call`` method until it's not rate-limited.

        The method is invoked while it returns 503 Service Unavailable or the
        allowed number of retries is reached.

        :param callable call: Method to be decorated.
        """

        def wrapper(*args, **kwargs):
            last_exception = None

            for _ in range(self.retries + 1):
                try:
                    return call(*args, **kwargs)
                except ServiceUnavailableError as e:
                    last_exception = e
                    time.sleep(self.sleep)  # hit by rate limit, let's sleep

            if last_exception:
                raise last_exception  # pylint: disable=raising-bad-type

        update_wrapper(wrapper, call)
        return wrapper


class VultrResponse(JsonResponse):
    def parse_error(self):
        if self.status == httplib.OK:
            body = self.parse_body()
            return body
        elif self.status == httplib.FORBIDDEN:
            raise InvalidCredsError(self.body)
        elif self.status == httplib.SERVICE_UNAVAILABLE:
            raise ServiceUnavailableError(self.body)
        else:
            raise LibcloudError(self.body)


class SSHKey(object):
    def __init__(self, id, name, pub_key):
        self.id = id
        self.name = name
        self.pub_key = pub_key

    def __repr__(self):
        return (('<SSHKey: id=%s, name=%s, pub_key=%s>') %
                (self.id, self.name, self.pub_key))


class VultrConnection(ConnectionKey):
    """
    Connection class for the Vultr driver.
    """

    host = 'api.vultr.com'
    responseCls = VultrResponse
    unauthenticated_endpoints = {  # {action: methods}
        '/v1/app/list': ['GET'],
        '/v1/os/list': ['GET'],
        '/v1/plans/list': ['GET'],
        '/v1/plans/list_vc2': ['GET'],
        '/v1/plans/list_vdc2': ['GET'],
        '/v1/regions/availability': ['GET'],
        '/v1/regions/list': ['GET']
    }

    def add_default_headers(self, headers):
        """
        Adds ``API-Key`` default header.

        :return: Updated headers.
        :rtype: dict
        """

        if self.require_api_key():
            headers.update({'API-Key': self.key})
        return headers

    def encode_data(self, data):
        return urlencode(data)

    @rate_limited()
    def get(self, url):
        return self.request(url)

    @rate_limited()
    def post(self, url, data):
        headers = {'Content-Type': 'application/x-www-form-urlencoded'}
        return self.request(url, data=data, headers=headers, method='POST')

    def require_api_key(self):
        """
        Check whether this call (method + action) must be authenticated.

        :return: True if ``API-Key`` header required, False otherwise.
        :rtype: bool
        """

        try:
            return self.method \
                not in self.unauthenticated_endpoints[self.action]
        except KeyError:
            return True


class VultrNodeDriverHelper(object):
    """
        VultrNode helper class.
    """

    def handle_extra(self, extra_keys, data):
        extra = {}
        for key in extra_keys:
            if key in data:
                extra[key] = data[key]
        return extra


class VultrNodeDriver(NodeDriver):
    type = Provider.VULTR
    name = 'Vultr'
    website = 'https://www.vultr.com'

    def __new__(cls, key, secret=None, secure=True, host=None, port=None,
                api_version=DEFAULT_API_VERSION, region=None, **kwargs):
        if cls is VultrNodeDriver:
            if api_version == '1':
                cls = VultrNodeDriverV1
            elif api_version == '2':
                cls = VultrNodeDriverV2
            else:
                raise NotImplementedError(
                    'No Vultr driver found for API version: %s' %
                    (api_version))
        return super().__new__(cls)


class VultrNodeDriverV1(VultrNodeDriver):
    """
    VultrNode node driver.
    """

    connectionCls = VultrConnection

    NODE_STATE_MAP = {'pending': NodeState.PENDING,
                      'active': NodeState.RUNNING}

    EX_CREATE_YES_NO_ATTRIBUTES = ['enable_ipv6',
                                   'enable_private_network',
                                   'auto_backups',
                                   'notify_activate',
                                   'ddos_protection']

    EX_CREATE_ID_ATTRIBUTES = {'iso_id': 'ISOID',
                               'script_id': 'SCRIPTID',
                               'snapshot_id': 'SNAPSHOTID',
                               'app_id': 'APPID'}

    EX_CREATE_ATTRIBUTES = ['ipxe_chain_url',
                            'label',
                            'userdata',
                            'reserved_ip_v4',
                            'hostname',
                            'tag']
    EX_CREATE_ATTRIBUTES.extend(EX_CREATE_YES_NO_ATTRIBUTES)
    EX_CREATE_ATTRIBUTES.extend(EX_CREATE_ID_ATTRIBUTES.keys())

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._helper = VultrNodeDriverHelper()

    def list_nodes(self):
        return self._list_resources('/v1/server/list', self._to_node)

    def list_key_pairs(self):
        """
        List all the available SSH keys.
        :return: Available SSH keys.
        :rtype: ``list`` of :class:`SSHKey`
        """
        return self._list_resources('/v1/sshkey/list', self._to_ssh_key)

    def create_key_pair(self, name, public_key=''):
        """
        Create a new SSH key.
        :param name: Name of the new SSH key
        :type name: ``str``

        :key public_key: Public part of the new SSH key
        :type name: ``str``

        :return: True on success
        :rtype: ``bool``
        """
        params = {'name': name, 'ssh_key': public_key}
        res = self.connection.post('/v1/sshkey/create', params)
        return res.status == httplib.OK

    def delete_key_pair(self, key_pair):
        """
        Delete an SSH key.
        :param key_pair: The SSH key to delete
        :type key_pair: :class:`SSHKey`

        :return: True on success
        :rtype: ``bool``
        """
        params = {'SSHKEYID': key_pair.id}
        res = self.connection.post('/v1/sshkey/destroy', params)
        return res.status == httplib.OK

    def list_locations(self):
        return self._list_resources('/v1/regions/list', self._to_location)

    def list_sizes(self):
        return self._list_resources('/v1/plans/list', self._to_size)

    def list_images(self):
        return self._list_resources('/v1/os/list', self._to_image)

    # pylint: disable=too-many-locals
    def create_node(self, name, size, image, location, ex_ssh_key_ids=None,
                    ex_create_attr=None):
        """
        Create a node

        :param name: Name for the new node
        :type name: ``str``

        :param size: Size of the new node
        :type size: :class:`NodeSize`

        :param image: Image for the new node
        :type image: :class:`NodeImage`

        :param location: Location of the new node
        :type location: :class:`NodeLocation`

        :param ex_ssh_key_ids: IDs of the SSH keys to initialize
        :type ex_sshkeyid: ``list`` of ``str``

        :param ex_create_attr: Extra attributes for node creation
        :type ex_create_attr: ``dict``

        The `ex_create_attr` parameter can include the following dictionary
        key and value pairs:

        * `ipxe_chain_url`: ``str`` for specifying URL to boot via IPXE
        * `iso_id`: ``str`` the ID of a specific ISO to mount,
          only meaningful with the `Custom` `NodeImage`
        * `script_id`: ``int`` ID of a startup script to execute on boot,
          only meaningful when the `NodeImage` is not `Custom`
        * 'snapshot_id`: ``str`` Snapshot ID to restore for the initial
          installation, only meaningful with the `Snapshot` `NodeImage`
        * `enable_ipv6`: ``bool`` Whether an IPv6 subnet should be assigned
        * `enable_private_network`: ``bool`` Whether private networking
          support should be added
        * `label`: ``str`` Text label to be shown in the control panel
        * `auto_backups`: ``bool`` Whether automatic backups should be enabled
        * `app_id`: ``int`` App ID to launch if launching an application,
          only meaningful when the `NodeImage` is `Application`
        * `userdata`: ``str`` Base64 encoded cloud-init user-data
        * `notify_activate`: ``bool`` Whether an activation email should be
          sent when the server is ready
        * `ddos_protection`: ``bool`` Whether DDOS protection should be enabled
        * `reserved_ip_v4`: ``str`` IP address of the floating IP to use as
          the main IP of this server
        * `hostname`: ``str`` The hostname to assign to this server
        * `tag`: ``str`` The tag to assign to this server

        :return: The newly created node.
        :rtype: :class:`Node`

        """
        params = {'DCID': location.id, 'VPSPLANID': size.id,
                  'OSID': image.id, 'label': name}

        if ex_ssh_key_ids is not None:
            params['SSHKEYID'] = ','.join(ex_ssh_key_ids)

        ex_create_attr = ex_create_attr or {}
        for key, value in ex_create_attr.items():
            if key in self.EX_CREATE_ATTRIBUTES:
                if key in self.EX_CREATE_YES_NO_ATTRIBUTES:
                    params[key] = 'yes' if value else 'no'
                else:
                    if key in self.EX_CREATE_ID_ATTRIBUTES:
                        key = self.EX_CREATE_ID_ATTRIBUTES[key]
                    params[key] = value

        result = self.connection.post('/v1/server/create', params)
        if result.status != httplib.OK:
            return False

        subid = result.object['SUBID']

        retry_count = 3
        created_node = None

        for _ in range(retry_count):
            try:
                nodes = self.list_nodes()
                created_node = [n for n in nodes if n.id == subid][0]
            except IndexError:
                time.sleep(1)
            else:
                break

        return created_node

    def reboot_node(self, node):
        params = {'SUBID': node.id}
        res = self.connection.post('/v1/server/reboot', params)

        return res.status == httplib.OK

    def destroy_node(self, node):
        params = {'SUBID': node.id}
        res = self.connection.post('/v1/server/destroy', params)

        return res.status == httplib.OK

    def _list_resources(self, url, tranform_func):
        data = self.connection.get(url).object
        sorted_key = sorted(data)
        return [tranform_func(data[key]) for key in sorted_key]

    def _to_node(self, data):
        if 'status' in data:
            state = self.NODE_STATE_MAP.get(data['status'], NodeState.UNKNOWN)
            if state == NodeState.RUNNING and \
                    data['power_status'] != 'running':
                state = NodeState.STOPPED
        else:
            state = NodeState.UNKNOWN

        if 'main_ip' in data and data['main_ip'] is not None:
            public_ips = [data['main_ip']]
        else:
            public_ips = []
        # simple check that we have ip address in value
        if len(data['internal_ip']) > 0:
            private_ips = [data['internal_ip']]
        else:
            private_ips = []
        created_at = parse_date(data['date_created'])

        # response ordering
        extra_keys = [
            "location",  # Location name
            "default_password", "pending_charges", "cost_per_month",
            "current_bandwidth_gb", "allowed_bandwidth_gb", "netmask_v4",
            "gateway_v4", "power_status", "server_state",
            "v6_networks",
            # TODO: Does we really need kvm_url?
            "kvm_url",
            "auto_backups", "tag",
            # "OSID",  # Operating system to use. See v1/os/list.
            "APPID", "FIREWALLGROUPID"
        ]
        extra = self._helper.handle_extra(extra_keys, data)

        resolve_data = VULTR_COMPUTE_INSTANCE_IMAGES.get(data["OSID"])
        if resolve_data:
            image = self._to_image(resolve_data)
        else:
            image = None

        resolve_data = VULTR_COMPUTE_INSTANCE_SIZES.get(data["VPSPLANID"])
        if resolve_data:
            size = self._to_size(resolve_data)
        else:
            size = None

        # resolve_data = VULTR_COMPUTE_INSTANCE_LOCATIONS.get(data['DCID'])
        # if resolve_data:
        #     location = self._to_location(resolve_data)
        # extra['location'] = location

        node = Node(
            id=data['SUBID'],
            name=data['label'],
            state=state,
            public_ips=public_ips,
            private_ips=private_ips,
            image=image,
            size=size,
            extra=extra,
            created_at=created_at,
            driver=self)

        return node

    def _to_location(self, data):
        extra_keys = ['continent', 'state', 'ddos_protection',
                      'block_storage', 'regioncode']
        extra = self._helper.handle_extra(extra_keys, data)

        return NodeLocation(id=data['DCID'], name=data['name'],
                            country=data['country'], extra=extra, driver=self)

    def _to_size(self, data):
        extra_keys = [
            'vcpu_count', 'plan_type', 'available_locations',
        ]
        extra = self._helper.handle_extra(extra_keys, data)

        # backward compatibility
        if extra.get('vcpu_count').isdigit():
            extra['vcpu_count'] = int(extra['vcpu_count'])

        ram = int(data['ram'])
        disk = int(data['disk'])
        # NodeSize accepted int instead float
        bandwidth = int(float(data['bandwidth']))
        price = float(data['price_per_month'])
        return NodeSize(id=data['VPSPLANID'], name=data['name'],
                        ram=ram, disk=disk,
                        bandwidth=bandwidth, price=price,
                        extra=extra, driver=self)

    def _to_image(self, data):
        extra_keys = ['arch', 'family']
        extra = self._helper.handle_extra(extra_keys, data)
        return NodeImage(id=data['OSID'], name=data['name'], extra=extra,
                         driver=self)

    def _to_ssh_key(self, data):
        return SSHKey(id=data['SSHKEYID'], name=data['name'],
                      pub_key=data['ssh_key'])


class VultrNodeDriverV2(VultrNodeDriver):
    """
    Vultr API v2 NodeDriver.
    """
    connectionCls = VultrConnectionV2
    NODE_STATE_MAP = {
        'active': NodeState.RUNNING,
        'halted': NodeState.STOPPED,
        'rebooting': NodeState.REBOOTING,
        'resizing': NodeState.RECONFIGURING,
        'pending': NodeState.PENDING,
    }

    VOLUME_STATE_MAP = {
        'active': StorageVolumeState.AVAILABLE,
        'pending': StorageVolumeState.CREATING,
    }

    SNAPSHOT_STATE_MAP = {
        'complete': VolumeSnapshotState.AVAILABLE,
        'pending': VolumeSnapshotState.CREATING,
    }

    def list_nodes(self, ex_list_bare_metals: bool = True) -> List[Node]:
        """List all nodes.

        :keyword ex_list_bare_metals: Whether to fetch bare metal nodes.
        :type    ex_list_bare_metals: ``bool``

        :return:  list of node objects
        :rtype: ``list`` of :class: `Node`
        """
        data = self._paginated_request('/v2/instances', 'instances')
        nodes = [self._to_node(item) for item in data]

        if ex_list_bare_metals:
            nodes += self.ex_list_bare_metal_nodes()
        return nodes

    def create_node(self,
                    name: str,
                    size: NodeSize,
                    location: NodeLocation,
                    image: Optional[NodeImage] = None,
                    ex_ssh_key_ids: Optional[List[str]] = None,
                    ex_private_network_ids: Optional[List[str]] = None,
                    ex_snapshot: Union[VultrNodeSnapshot, str, None] = None,
                    ex_enable_ipv6: bool = False,
                    ex_backups: bool = False,
                    ex_userdata: Optional[str] = None,
                    ex_ddos_protection: bool = False,
                    ex_enable_private_network: bool = False,
                    ex_ipxe_chain_url: Optional[str] = None,
                    ex_iso_id: Optional[str] = None,
                    ex_script_id: Optional[str] = None,
                    ex_image_id: Optional[str] = None,
                    ex_activation_email: bool = False,
                    ex_hostname: Optional[str] = None,
                    ex_tag: Optional[str] = None,
                    ex_firewall_group_id: Optional[str] = None,
                    ex_reserved_ipv4: Optional[str] = None,
                    ex_persistent_pxe: bool = False
                    ) -> Node:
        """Create a new node.

        :param name: The new node's name.
        :type name: ``str``

        :param size: The size to use to create the node.
        :type size: :class: `NodeSize`

        :param location: The location to provision the node.
        :type location: :class: `NodeLocation`

        :keyword image:  The image to use to provision the node.
        :type    image:  :class: `NodeImage`

        :keyword ex_ssh_key_ids: List of SSH keys to install on this node.
        :type    ex_ssh_key_ids: ``list`` of ``str``

        :keyword ex_private_network_ids: The network ids to attach to node.
                                         This parameter takes precedence over
                                         ex_enable_private_network (VPS only)
        :type    ex_private_network_ids: ``list`` of ``str``

        :keyword ex_snapshot: The snapshot to use when deploying the node.
                                 Mutually exclusive with image,
        :type    ex_snapshot: :class: `VultrNodeSnapshot` or ``str``

        :keyword ex_enable_ipv6: Wheteher to enable IPv6.
        :type    ex_enable_ipv6: ``bool``

        :keyword ex_backups: Enable automatic backups for the node. (VPS only)
        :type    ex_backups: ``bool``

        :keyword ex_userdata: String containing user data
        :type    ex_userdata: ``str``

        :keyword ex_ddos_protection: Enable DDoS protection (VPS only)
        :type    ex_ddos_protection: ``bool``

        :keyword ex_enable_private_network: Enable private networking.
                                            Mutually exclusive with
                                             ex_private_network_ids.
                                            (VPS only)
        :type    ex_enable_private_network: ``bool``

        :keyword ex_ipxe_chain_url: The URL location of the iPXE chainloader
                                    (VPS only)
        :type    ex_ipxe_chain_url: ``str``

        :keyword ex_iso_id: The ISO id to use when deploying this node.
                            (VPS only)
        :type    ex_iso_id: ``str``

        :keyword ex_script_id: The startup script id to use when deploying
                               this node.
        :type    ex_script_id: ``str``

        :keyword ex_image_id: The Application image_id to use when deploying
                              this node.
        :type    ex_image_id: ``str``

        :keyword ex_activation_email: Notify by email after deployment.
        :type    ex_activation_email: ``bool``

        :keyword ex_hostname: The hostname to use when deploying this node.
        :type    ex_hostname: ``str``

        :keyword ex_tag: The user-supplied tag.
        :type    ex_tag: ``str``

        :keyword ex_firewall_group_id: The Firewall Group id to attach to
                                       this node. (VPS only)
        :type    ex_firewall_group_id: ``str``

        :keyword ex_reserved_ipv4: Id of the floating IP to use as the
                                   main IP of this node.
        :type    ex_reserved_ipv4: ``str``

        :keyword ex_persistent_pxe: Enable persistent PXE (Bare Metal only)
        :type    ex_persistent_pxe: ``bool``
        """
        data = {
            'label': name,
            'region': location.id,
            'plan': size.id,
            'enable_ipv6': ex_enable_ipv6,
            'activation_email': ex_activation_email,
        }

        if image:
            data['os_id'] = image.id

        if ex_ssh_key_ids:
            data['sshkey_id'] = ex_ssh_key_ids

        if ex_snapshot:
            try:
                data['snapshot_id'] = ex_snapshot.id
            except AttributeError:
                data['snapshot_id'] = ex_snapshot

        if ex_userdata:
            data['user_data'] = base64.b64encode(
                bytes(ex_userdata, 'utf-8')).decode('utf-8')

        if ex_script_id:
            data['script_id'] = ex_script_id

        if ex_image_id:
            data['image_id'] = ex_image_id

        if ex_hostname:
            data['hostname'] = ex_hostname

        if ex_reserved_ipv4:
            data['reserved_ipv4'] = ex_reserved_ipv4

        if ex_tag:
            data['tag'] = ex_tag

        if self._is_bare_metal(size):
            if ex_persistent_pxe:
                data['persistent_pxe'] = ex_persistent_pxe
            resp = self.connection.request('/v2/bare-metals',
                                           data=json.dumps(data),
                                           method='POST')
            return self._to_node(resp.object['bare_metal'])
        else:
            if ex_private_network_ids:
                data['attach_private_network'] = ex_private_network_ids

            if ex_enable_private_network:
                data['enable_private_network'] = ex_enable_private_network

            if ex_ipxe_chain_url:
                data['ipxe_chain_url'] = ex_ipxe_chain_url

            if ex_iso_id:
                data['iso_id'] = ex_iso_id

            if ex_ddos_protection:
                data['ddos_protection'] = ex_ddos_protection

            if ex_firewall_group_id:
                data['firewall_group_id'] = ex_firewall_group_id

            if ex_backups:
                data['backups'] = ('enabled' if ex_backups is True
                                   else 'disabled')

            resp = self.connection.request('/v2/instances',
                                           data=json.dumps(data),
                                           method='POST')
            return self._to_node(resp.object['instance'])

    def reboot_node(self, node: Node) -> bool:
        """Reboot the given node.

        :param node: The node to be rebooted.
        :type node: :class: `Node`

        :rtype: ``bool``
        """
        if self._is_bare_metal(node.size):
            return self.ex_reboot_bare_metal_node(node)

        resp = self.connection.request('/v2/instances/%s/reboot' % node.id,
                                       method='POST')

        return resp.success()

    def start_node(self, node: Node) -> bool:
        """Start the given node.

        :param node: The node to be started.
        :type node: :class: `Node`

        :rtype: ``bool``
        """
        if self._is_bare_metal(node.size):
            return self.ex_start_bare_metal_node(node)

        resp = self.connection.request('/v2/instances/%s/start' % node.id,
                                       method='POST')

        return resp.success()

    def stop_node(self, node: Node) -> bool:
        """Stop the given node.

        :param node: The node to be stopped.
        :type node: :class: `Node`

        :rtype: ``bool``
        """
        if self._is_bare_metal(node.size):
            return self.ex_stop_bare_metal_node(node)

        return self.ex_stop_nodes([node])

    def destroy_node(self, node: Node) -> bool:
        """Destroy the given node.

        :param node: The node to be destroyed.
        :type node: :class: `Node`

        :rtype: ``bool``
        """
        if self._is_bare_metal(node.size):
            return self.ex_destroy_bare_metal_node(node)

        resp = self.connection.request('/v2/instances/%s' % node.id,
                                       method='DELETE')

        return resp.success()

    def list_sizes(self, ex_list_bare_metals: bool = True) -> List[NodeSize]:
        """List available node sizes.

        :keyword ex_list_bare_metals: Whether to fetch bare metal sizes.
        :type    ex_list_bare_metals: ``bool``

        :rtype: ``list`` of :class: `NodeSize`
        """
        data = self._paginated_request('/v2/plans', 'plans')
        sizes = [self._to_size(item) for item in data]

        if ex_list_bare_metals:
            sizes += self.ex_list_bare_metal_sizes()
        return sizes

    def list_images(self) -> List[NodeImage]:
        """List available node images.

        :rtype: ``list`` of :class: `NodeImage`
        """
        data = self._paginated_request('/v2/os', 'os')
        return [self._to_image(item) for item in data]

    def list_locations(self) -> List[NodeLocation]:
        """List available node locations.

        :rtype: ``list`` of :class: `NodeLocation`
        """
        data = self._paginated_request('/v2/regions', 'regions')
        return [self._to_location(item) for item in data]

    def list_volumes(self) -> List[StorageVolume]:
        """List storage volumes.

        :rtype: ``list`` of :class:`StorageVolume`
        """
        data = self._paginated_request('/v2/blocks', 'blocks')
        return [self._to_volume(item) for item in data]

    def create_volume(self,
                      size: int,
                      name: str,
                      location: Union[NodeLocation, str],
                      ) -> StorageVolume:
        """Create a new volume.

        :param size: Size of the volume in gigabytes.\
        Size may range between 10 and 10000.
        :type size: ``int``

        :param name: Name of the volume to be created.
        :type name: ``str``

        :param location: Which data center to create the volume in.
        :type location: :class:`NodeLocation` or ``str``

        :return: The newly created volume.
        :rtype: :class:`StorageVolume`
        """

        data = {
            'label': name,
            'size_gb': size,
        }
        try:
            data['region'] = location.id
        except AttributeError:
            data['region'] = location

        resp = self.connection.request('/v2/blocks',
                                       data=json.dumps(data),
                                       method='POST')
        return self._to_volume(resp.object['block'])

    def attach_volume(self,
                      node: Node,
                      volume: StorageVolume,
                      ex_live: bool = True,
                      ) -> bool:
        """Attaches volume to node.

        :param node: Node to attach volume to.
        :type node: :class:`Node`

        :param volume: Volume to attach.
        :type volume: :class:`StorageVolume`

        :param ex_live: Attach the volume without restarting the node.
        :type ex_live: ``bool``

        :rytpe: ``bool``
        """

        data = {
            'instance_id': node.id,
            'live': ex_live,
        }
        resp = self.connection.request('/v2/blocks/%s/attach' % volume.id,
                                       data=json.dumps(data),
                                       method='POST')

        return resp.success()

    def detach_volume(self,
                      volume: StorageVolume,
                      ex_live: bool = True,
                      ) -> bool:
        """Detaches a volume from a node.

        :param volume: Volume to be detached
        :type volume: :class:`StorageVolume`

        :param ex_live: Detach the volume without restarting the node.
        :type ex_live: ``bool``

        :rtype: ``bool``
        """
        data = {
            'live': ex_live
        }

        resp = self.connection.request('/v2/blocks/%s/detach' % volume.id,
                                       data=json.dumps(data),
                                       method='POST')

        return resp.success()

    def destroy_volume(self, volume: StorageVolume) -> bool:
        """Destroys a storage volume.

        :param volume: Volume to be destroyed
        :type  volume: :class:`StorageVolume`

        :rtype: ``bool``
        """

        resp = self.connection.request('/v2/blocks/%s' % volume.id,
                                       method='DELETE')

        return resp.success()

    def list_key_pairs(self) -> List[KeyPair]:
        """List all the available SSH key pair objects.

        :rtype: ``list`` of :class:`KeyPair`
        """
        data = self._paginated_request('/v2/ssh-keys', 'ssh_keys')
        return [self._to_key_pair(item) for item in data]

    def get_key_pair(self, key_id: str) -> KeyPair:
        """Retrieve a single key pair.

        :param key_id: ID of the key pair to retrieve.
        :type  key_id: ``str``

        :rtype: :class: `KeyPair`
        """
        resp = self.connection.request('/v2/ssh-keys/%s' % key_id)
        return self._to_key_pair(resp.object['ssh_key'])

    def import_key_pair_from_string(self,
                                    name: str,
                                    key_material: str
                                    ) -> KeyPair:
        """Import a new public key from string.

        :param name: Key pair name.
        :type  name: ``str``

        :param key_material: Public key material.
        :type  key_material: ``str``

        :rtype: :class: `KeyPair`
        """
        data = {
            'name': name,
            'ssh_key': key_material,
        }
        resp = self.connection.request('/v2/ssh-keys',
                                       data=json.dumps(data),
                                       method='POST')
        return self._to_key_pair(resp.object['ssh_key'])

    def delete_key_pair(self, key_pair: KeyPair) -> bool:
        """Delete existing key pair.

        :param key_pair: The key pair object to delete.
        :type key_pair: :class:`.KeyPair`

        :rtype: ``bool``
        """

        resp = self.connection.request(
            '/v2/ssh-keys/%s' % key_pair.extra['id'],
            method='DELETE')

        return resp.success()

    def ex_list_bare_metal_nodes(self) -> List[Node]:
        """List all bare metal nodes.

        :return:  list of node objects
        :rtype: ``list`` of :class: `Node`
        """
        data = self._paginated_request('/v2/bare-metals', 'bare_metals')
        return [self._to_node(item) for item in data]

    def ex_reboot_bare_metal_node(self, node: Node) -> bool:
        """Reboot the given bare metal node.

        :param node: The bare metal node to be rebooted.
        :type node: :class: `Node`

        :rtype: ``bool``
        """
        resp = self.connection.request('/v2/bare-metals/%s/reboot' % node.id,
                                       method='POST')

        return resp.success()

    def ex_resize_node(self, node: Node, size: NodeSize) -> bool:
        """Change size for the given node, only applicable for VPS nodes.

        :param node: The node to be resized.
        :type node: :class: `Node`

        :param size: The new size.
        :type size: :class: `NodeSize`
        """
        data = {
            'plan': size.id
        }
        resp = self.connection.request('/v2/instances/%s' % node.id,
                                       data=json.dumps(data),
                                       method='PATCH')
        return self._to_node(resp.object['instance'])

    def ex_start_bare_metal_node(self, node: Node) -> bool:
        """Start the given bare metal node.

        :param node: The bare metal node to be started.
        :type node: :class: `Node`

        :rtype: ``bool``
        """
        resp = self.connection.request('/v2/bare-metals/%s/start' % node.id,
                                       method='POST')

        return resp.success()

    def ex_stop_bare_metal_node(self, node: Node) -> bool:
        """Stop the given bare metal node.

        :param node: The bare metal node to be stopped.
        :type node: :class: `Node`

        :rtype: ``bool``
        """
        resp = self.connection.request('/v2/bare-metals/%s/halt' % node.id,
                                       method='POST')

        return resp.success()

    def ex_destroy_bare_metal_node(self, node: Node) -> bool:
        """Destroy the given bare metal node.

        :param node: The bare metal node to be destroyed.
        :type node: :class: `Node`

        :rtype: ``bool``
        """
        resp = self.connection.request('/v2/bare-metals/%s' % node.id,
                                       method='DELETE')

        return resp.success()

    def ex_get_node(self, node_id: str) -> Node:
        """Retrieve a node object.

        :param node_id: ID of the node to retrieve.
        :type snapshot_id: ``str``

        :rtype: :class: `Node`
        """
        resp = self.connection.request('/v2/instances/%s' % node_id)
        return self._to_node(resp.object['instance'])

    def ex_stop_nodes(self, nodes: List[Node]) -> bool:
        """Stops all the nodes given.

        : param nodes: A list of the nodes to stop.
        : type node: ``list`` of: class `Node`

        : rtype: ``bool``
        """

        data = {
            "instance_ids": [node.id for node in nodes]
        }
        resp = self.connection.request('/v2/instances/halt',
                                       data=json.dumps(data),
                                       method='POST')

        return resp.success()

    def ex_list_bare_metal_sizes(self) -> List[NodeSize]:
        """List bare metal sizes.

        :rtype: ``list`` of :class: `NodeSize`
        """
        data = self._paginated_request('/v2/plans-metal', 'plans_metal')
        return [self._to_size(item) for item in data]

    def ex_list_snapshots(self) -> List[VultrNodeSnapshot]:
        """List node snapshots.

        :rtype: ``list`` of :class: `VultrNodeSnapshot`
        """
        data = self._paginated_request('/v2/snapshots', 'snapshots')
        return [self._to_snapshot(item) for item in data]

    def ex_get_snapshot(self, snapshot_id: str) -> VultrNodeSnapshot:
        """Retrieve a snapshot.

        :param snapshot_id: ID of the snapshot to retrieve.
        :type snapshot_id: ``str``

        :rtype: :class: `VultrNodeSnapshot`
        """
        resp = self.connection.request('/v2/snapshots/%s' % snapshot_id)
        return self._to_snapshot(resp.object['snapshot'])

    def ex_create_snapshot(self,
                           node: Node,
                           description: Optional[str] = None
                           ) -> VultrNodeSnapshot:
        """Create snapshot from a node.

        :param node: Node to create the snapshot from.
        :type node: :class: `Node`

        :keyword description: A description of the snapshot.
        :type    description: ``str``

        :rtype: :class: `VultrNodeSnapshot`
        """
        data = {
            'instance_id': node.id,
        }
        if description:
            data['description'] = description

        resp = self.connection.request('/v2/snapshots',
                                       data=json.dumps(data),
                                       method='POST')

        return self._to_snapshot(resp.object['snapshot'])

    def ex_delete_snapshot(self, snapshot: VultrNodeSnapshot) -> bool:
        """Delete the given snapshot.

        :param snapshot: The snapshot to delete.
        :type node: :class:`VultrNodeSnapshot`

        :rtype: ``bool``
        """

        resp = self.connection.request('/v2/snapshots/%s' % snapshot.id,
                                       method='DELETE')

        return resp.success()

    def ex_list_networks(self) -> List[VultrNetwork]:
        """List all private networks.

        :rtype: ``list`` of :class: `VultrNetwork`
        """

        data = self._paginated_request('/v2/private-networks', 'networks')
        return [self._to_network(item) for item in data]

    def ex_create_network(self,
                          cidr_block: str,
                          location: Union[NodeLocation, str],
                          description: Optional[str] = None,
                          ) -> VultrNetwork:
        """Create a private network.

        :param cidr_block: The CIDR block assigned to the network.
        :type  cidr_block: ``str``

        :param location: The location to create the network.
        :type  location: :class: `NodeLocation` or ``str``

        :keyword description: A description of the private network.
        :type    description: ``str``

        :rtype: :class: `VultrNetwork`
        """
        subnet, subnet_mask = cidr_block.split('/')
        data = {
            'v4_subnet': subnet,
            'v4_subnet_mask': int(subnet_mask),
        }

        try:
            data['region'] = location.id
        except AttributeError:
            data['region'] = location

        if description:
            data['description'] = description

        resp = self.connection.request('/v2/private-networks',
                                       data=json.dumps(data),
                                       method='POST')

        return self._to_network(resp.object['network'])

    def ex_get_network(self, network_id: str) -> VultrNetwork:
        """Retrieve a private network.

        :param network_id: ID of the network to retrieve.
        :type network_id: ``str``

        :rtype: :class: `VultrNetwork`
        """

        resp = self.connection.request('/v2/private-networks/%s' % network_id)
        return self._to_network(resp.object['network'])

    def ex_destroy_network(self, network: VultrNetwork) -> bool:
        """Delete a private network.

        :param network: The network to destroy.
        :type  network: :class: `VultrNetwork`

        :rtype: ``bool``
        """
        resp = self.connection.request('/v2/private-networks/%s' % network.id,
                                       method='DELETE')

        return resp.success()

    def ex_list_available_sizes_for_location(self,
                                             location: NodeLocation,
                                             ) -> List[str]:
        """Get a list of available sizes for the given location.

        :param location: The location to get available sizes for.
        :type location: :class: `NodeLocation`

        :return:  A list of available size IDs for the given location.
        :rtype: ``list`` of ``str``
        """
        resp = self.connection.request('/v2/regions/%s/availability'
                                       % location.id)
        return resp.object['available_plans']

    def ex_get_volume(self, volume_id: str) -> StorageVolume:
        """Retrieve a single volume.

        :param volume_id: The ID of the volume to fetch.
        :type volume_id: ``str``

        :rtype :class: `StorageVolume`
        :return: StorageVolume instance on success.
        """
        resp = self.connection.request('/v2/blocks/%s' % volume_id)

        return self._to_volume(resp.object['block'])

    def ex_resize_volume(self, volume: StorageVolume, size: int) -> bool:
        """Resize a volume.

        :param volume: The volume to resize.
        :type volume: :class:`StorageVolume`

        :param size: The new volume size in GBs.\
        Size may range between 10 and 10000.
        :type size: ``int``

        :rtype: ``bool``
        """
        data = {
            'label': volume.name,
            'size_gb': size,
        }

        resp = self.connection.request('/v2/blocks/%s' % volume.id,
                                       data=json.dumps(data),
                                       method='PATCH')
        return resp.success()

    def _is_bare_metal(self, size: Union[NodeSize, str]) -> bool:
        try:
            size_id = size.id
        except AttributeError:
            size_id = size

        return size_id.startswith('vbm')

    def _to_node(self, data: Dict[str, Any]) -> Node:
        id_ = data['id']
        name = data['label']
        public_ips = data['main_ip'].split() + data['v6_main_ip'].split()
        size = data['plan']
        image = str(data['os_id'])
        created_at = data['date_created']
        is_bare_metal = self._is_bare_metal(size)
        extra = {
            'location': data['region'],
            'ram': data['ram'],
            'disk': data['disk'],
            'netmask_v4': data['netmask_v4'],
            'gateway_v4': data['gateway_v4'],
            'v6_network': data['v6_network'],
            'v6_network_size': data['v6_network_size'],
            'app_id': data['app_id'],
            'image_id': data['image_id'],
            'features': data['features'],
            'tag': data['tag'],
            'os': data['os'],
            'is_bare_metal': is_bare_metal,
        }
        if is_bare_metal:
            state = self._get_node_state(data['status'])
            extra['cpu_count'] = data['cpu_count']
            extra['mac_address'] = data['mac_address']
            private_ips = None
        else:
            state = self._get_node_state(data['status'],
                                         power_state=data['power_status'])
            extra['vcpu_count'] = data['vcpu_count']
            extra['allowed_bandwidth'] = data['allowed_bandwidth']
            extra['power_status'] = data['power_status']
            extra['server_status'] = data['server_status']
            extra['firewall_group_id'] = data['firewall_group_id']
            private_ips = data['internal_ip'].split()

        return Node(id=id_,
                    name=name,
                    state=state,
                    public_ips=public_ips,
                    private_ips=private_ips,
                    driver=self,
                    size=size,
                    image=image,
                    extra=extra,
                    created_at=created_at)

    def _to_volume(self, data: Dict[str, Any]) -> StorageVolume:
        id_ = data['id']
        name = data['label']
        size = data['size_gb']
        try:
            state = self.VOLUME_STATE_MAP[data['status']]
        except KeyError:
            state = StorageVolumeState.UNKNOWN
        extra = {
            'date_created': data['date_created'],
            'cost': data['cost'],
            'location': data['region'],
            'attached_to_instance': data['attached_to_instance'],
            'mount_id': data['mount_id'],
        }
        return StorageVolume(id=id_,
                             name=name,
                             size=size,
                             driver=self,
                             state=state,
                             extra=extra)

    def _get_node_state(self,
                        state: str,
                        power_state: Optional[str] = None,
                        ) -> NodeState:
        try:
            state = self.NODE_STATE_MAP[state]
        except KeyError:
            state = NodeState.UNKNOWN

        if power_state is None:
            return state

        if state == NodeState.RUNNING and power_state != 'running':
            state = NodeState.STOPPED
        return state

    def _to_key_pair(self, data: Dict[str, Any]) -> KeyPair:
        name = data['name']
        public_key = data['ssh_key']
        # requires cryptography module
        try:
            fingerprint = get_pubkey_openssh_fingerprint(public_key)
        except RuntimeError:
            fingerprint = None
        extra = {
            'id': data['id'],
            'date_created': data['date_created'],
        }
        return KeyPair(name=name,
                       public_key=public_key,
                       fingerprint=fingerprint,
                       driver=self,
                       extra=extra,
                       )

    def _to_location(self, data: Dict[str, Any]) -> NodeLocation:
        id_ = data['id']
        name = data['city']
        country = data['country']
        extra = {
            'continent': data['continent'],
            'option': data['options'],
        }
        return NodeLocation(id=id_,
                            name=name,
                            country=country,
                            driver=self,
                            extra=extra
                            )

    def _to_image(self, data: Dict[str, Any]) -> NodeImage:
        id_ = data['id']
        name = data['name']
        extra = {
            'arch': data['arch'],
            'family': data['family'],
        }
        return NodeImage(id=id_,
                         name=name,
                         driver=self,
                         extra=extra)

    def _to_size(self, data: Dict[str, Any]) -> NodeSize:
        id_ = data['id']
        ram = data['ram']
        disk = data['disk']
        bandwidth = data['bandwidth']
        price = data['monthly_cost']
        is_bare_metal = self._is_bare_metal(id_)
        extra = {
            'locations': data['locations'],
            'type': data['type'],
            'disk_count': data['disk_count'],
            'is_bare_metal': is_bare_metal,
        }

        # VPS and bare metal sizes have different fields
        if is_bare_metal is False:
            extra['vcpu_count'] = data['vcpu_count']
        else:
            extra['cpu_count'] = data['cpu_count']
            extra['cpu_model'] = data['cpu_model']
            extra['cpu_threads'] = data['cpu_threads']

        return NodeSize(id=id_,
                        name=id_,
                        ram=ram,
                        disk=disk,
                        bandwidth=bandwidth,
                        price=price,
                        driver=self,
                        extra=extra)

    def _to_network(self, data: Dict[str, Any]) -> VultrNetwork:
        id_ = data['id']
        cidr_block = '%s/%s' % (data['v4_subnet'], data['v4_subnet_mask'])
        location = data['region']
        extra = {
            'description': data['description'],
            'date_created': data['date_created'],
        }
        return VultrNetwork(id=id_,
                            cidr_block=cidr_block,
                            location=location,
                            extra=extra)

    def _to_snapshot(self, data: Dict[str, Any]) -> VultrNodeSnapshot:
        id_ = data['id']
        created = data['date_created']
        # Size is returned in bytes, convert to GBs
        size = data['size'] / 1024 / 1024 / 1024
        try:
            state = self.SNAPSHOT_STATE_MAP[data['status']]
        except KeyError:
            state = VolumeSnapshotState.UNKNOWN
        extra = {
            'description': data['description'],
            'os_id': data['os_id'],
            'app_id': data['app_id'],
        }
        return VultrNodeSnapshot(id=id_,
                                 size=size,
                                 created=created,
                                 state=state,
                                 extra=extra,
                                 driver=self)

    def _paginated_request(self,
                           url: str,
                           key: str,
                           params: Optional[Dict[str, Any]] = None
                           ) -> List[Any]:
        """Perform multiple calls to get the full list of items when
        the API responses are paginated.

        :param url: API endpoint
        :type url: ``str``

        :param key: Result object key
        :type key: ``str``

        :param params: Request parameters
        :type params: ``dict``

        :return: ``list`` of API response objects
        :rtype: ``list``
        """
        params = params if params is not None else {}
        resp = self.connection.request(url, params=params).object
        data = list(resp.get(key, []))
        objects = data
        while True:
            next_page = resp['meta']['links']['next']
            if next_page:
                params['cursor'] = next_page
                resp = self.connection.request(url, params=params).object
                data = list(resp.get(key, []))
                objects.extend(data)
            else:
                return objects



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and

import sys
import json

from libcloud.common.types import ProviderError
from libcloud.dns.drivers.auroradns import AuroraDNSDriver
from libcloud.dns.drivers.auroradns import AuroraDNSHealthCheckType
from libcloud.dns.types import RecordType
from libcloud.dns.types import ZoneDoesNotExistError
from libcloud.dns.types import ZoneAlreadyExistsError
from libcloud.dns.types import RecordDoesNotExistError
from libcloud.dns.base import Zone
from libcloud.test import LibcloudTestCase
from libcloud.test import MockHttp
from libcloud.test import unittest
from libcloud.test.file_fixtures import DNSFileFixtures
from libcloud.test.secrets import DNS_PARAMS_AURORADNS
from libcloud.utils.py3 import httplib


class AuroraDNSDriverTests(LibcloudTestCase):

    def setUp(self):
        AuroraDNSDriver.connectionCls.conn_class = AuroraDNSDriverMockHttp
        AuroraDNSDriverMockHttp.type = None
        self.driver = AuroraDNSDriver(*DNS_PARAMS_AURORADNS)

    def test_403_status_code(self):
        AuroraDNSDriverMockHttp.type = "HTTP_FORBIDDEN"

        with self.assertRaises(ProviderError) as ctx:
            self.driver.list_zones()

        self.assertEqual(ctx.exception.value, "Authorization failed")
        self.assertEqual(ctx.exception.http_code, 403)

    def test_merge_extra_data(self):
        rdata = {
            'name': 'localhost',
            'type': RecordType.A,
            'content': '127.0.0.1'
        }

        params = {'ttl': 900,
                  'prio': 0,
                  'health_check_id': None,
                  'disabled': False}

        for param in params:
            extra = {
                param: params[param]
            }

            data = self.driver._AuroraDNSDriver__merge_extra_data(rdata, extra)
            self.assertEqual(data['content'], '127.0.0.1')
            self.assertEqual(data['type'], RecordType.A)
            self.assertEqual(data[param], params[param])
            self.assertEqual(data['name'], 'localhost')

    def test_res_to_record(self):
        res = {'id': 2,
               'name': 'www',
               'type': 'AAAA',
               'content': '2001:db8:100',
               'created': 1234,
               'modified': 2345,
               'disabled': False,
               'ttl': 1800,
               'prio': 10}

        zone = Zone(id=1,
                    domain='example.com',
                    type=None,
                    ttl=60,
                    driver=self.driver)

        record = self.driver._AuroraDNSDriver__res_to_record(zone, res)
        self.assertEqual(res['name'], record.name)
        self.assertEqual(res['ttl'], record.extra['ttl'])
        self.assertEqual(res['prio'], record.extra['priority'])
        self.assertEqual(res['type'], record.type)
        self.assertEqual(res['content'], record.data)
        self.assertEqual(zone, record.zone)
        self.assertEqual(self.driver, record.driver)

    def test_record_types(self):
        types = self.driver.list_record_types()
        self.assertEqual(len(types), 12)
        self.assertTrue(RecordType.A in types)
        self.assertTrue(RecordType.AAAA in types)
        self.assertTrue(RecordType.MX in types)
        self.assertTrue(RecordType.NS in types)
        self.assertTrue(RecordType.SOA in types)
        self.assertTrue(RecordType.TXT in types)
        self.assertTrue(RecordType.CNAME in types)
        self.assertTrue(RecordType.SRV in types)
        self.assertTrue(RecordType.DS in types)
        self.assertTrue(RecordType.SSHFP in types)
        self.assertTrue(RecordType.PTR in types)
        self.assertTrue(RecordType.TLSA in types)

    def test_list_zones(self):
        zones = self.driver.list_zones()
        self.assertEqual(len(zones), 2)
        for zone in zones:
            self.assertTrue(zone.domain.startswith('auroradns'))

    def test_create_zone(self):
        zone = self.driver.create_zone('example.com')
        self.assertEqual(zone.domain, 'example.com')

    def test_get_zone(self):
        zone = self.driver.get_zone('example.com')
        self.assertEqual(zone.domain, 'example.com')
        self.assertEqual(zone.id, 'ffb62570-8414-4578-a346-526b44e320b7')

    def test_delete_zone(self):
        zone = self.driver.get_zone('example.com')
        self.assertTrue(self.driver.delete_zone(zone))

    def test_create_record(self):
        zone = self.driver.get_zone('example.com')
        record = zone.create_record(name='localhost',
                                    type=RecordType.A,
                                    data='127.0.0.1',
                                    extra={'ttl': 900})
        self.assertEqual(record.id, '5592f1ff')
        self.assertEqual(record.name, 'localhost')
        self.assertEqual(record.data, '127.0.0.1')
        self.assertEqual(record.type, RecordType.A)
        self.assertEqual(record.extra['ttl'], 900)

    def test_get_record(self):
        zone = self.driver.get_zone('example.com')
        record = self.driver.get_record(zone.id, '5592f1ff')
        self.assertEqual(record.id, '5592f1ff')
        self.assertEqual(record.name, 'localhost')
        self.assertEqual(record.data, '127.0.0.1')
        self.assertEqual(record.type, RecordType.A)
        self.assertEqual(record.extra['ttl'], 900)
        self.assertEqual(record.extra['priority'], None)

    def test_update_record(self):
        ttl = 900
        zone = self.driver.get_zone('example.com')
        record = self.driver.get_record(zone.id, '5592f1ff')
        record = record.update(extra={'ttl': ttl})
        self.assertEqual(record.extra['ttl'], ttl)

    def test_delete_record(self):
        zone = self.driver.get_zone('example.com')
        record = self.driver.get_record(zone.id, '5592f1ff')
        self.assertTrue(record.delete())

    def test_list_records(self):
        zone = self.driver.get_zone('example.com')
        for record in zone.list_records():
            self.assertEqual(record.extra['ttl'], 3600)
            self.assertEqual(record.extra['disabled'], False)

    def test_get_zone_non_exist(self):
        try:
            self.driver.get_zone('nonexists.example.com')
            self.fail('expected a ZoneDoesNotExistError')
        except ZoneDoesNotExistError:
            pass
        except Exception:
            raise

    def test_delete_zone_non_exist(self):
        try:
            self.driver.delete_zone(Zone(id=1, domain='nonexists.example.com',
                                         type='NATIVE', driver=AuroraDNSDriver,
                                         ttl=3600))
            self.fail('expected a ZoneDoesNotExistError')
        except ZoneDoesNotExistError:
            pass
        except Exception:
            raise

    def test_create_zone_already_exist(self):
        try:
            self.driver.create_zone('exists.example.com')
            self.fail('expected a ZoneAlreadyExistsError')
        except ZoneAlreadyExistsError:
            pass
        except Exception:
            raise

    def test_list_records_non_exist(self):
        try:
            self.driver.list_records(Zone(id=1, domain='nonexists.example.com',
                                          type='NATIVE', driver=AuroraDNSDriver,
                                          ttl=3600))
            self.fail('expected a ZoneDoesNotExistError')
        except ZoneDoesNotExistError:
            pass
        except Exception:
            raise

    def test_get_record_non_exist(self):
        try:
            self.driver.get_record(1, 1)
            self.fail('expected a RecordDoesNotExistError')
        except RecordDoesNotExistError:
            pass
        except Exception:
            raise

    def test_create_health_check(self):
        zone = self.driver.get_zone('example.com')

        type = AuroraDNSHealthCheckType.HTTP
        hostname = "www.pcextreme.nl"
        ipaddress = "109.72.87.252"
        port = 8080
        interval = 10
        threshold = 3

        check = self.driver.ex_create_healthcheck(zone=zone,
                                                  type=type,
                                                  hostname=hostname,
                                                  port=port,
                                                  path=None,
                                                  interval=interval,
                                                  threshold=threshold,
                                                  ipaddress=ipaddress)

        self.assertEqual(check.interval, interval)
        self.assertEqual(check.threshold, threshold)
        self.assertEqual(check.port, port)
        self.assertEqual(check.type, type)
        self.assertEqual(check.hostname, hostname)
        self.assertEqual(check.path, "/")
        self.assertEqual(check.ipaddress, ipaddress)

    def test_list_health_checks(self):
        zone = self.driver.get_zone('example.com')
        checks = self.driver.ex_list_healthchecks(zone)

        self.assertEqual(len(checks), 3)

        for check in checks:
            self.assertEqual(check.interval, 60)
            self.assertEqual(check.type, AuroraDNSHealthCheckType.HTTP)


class AuroraDNSDriverMockHttp(MockHttp):
    fixtures = DNSFileFixtures('auroradns')

    def _zones(self, method, url, body, headers):
        if method == 'POST':
            body_json = json.loads(body)
            if body_json['name'] == 'exists.example.com':
                return (httplib.CONFLICT, body, {},
                        httplib.responses[httplib.CONFLICT])
            body = self.fixtures.load('zone_example_com.json')
        else:
            body = self.fixtures.load('zone_list.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _zones_HTTP_FORBIDDEN(self, method, url, body, headers):
        body = "{}"
        return (httplib.FORBIDDEN, body, {}, httplib.responses[httplib.FORBIDDEN])

    def _zones_example_com(self, method, url, body, headers):
        body = None
        if method == 'GET':
            body = self.fixtures.load('zone_example_com.json')

        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _zones_nonexists_example_com(self, method, url, body, headers):
        return (httplib.NOT_FOUND, body, {},
                httplib.responses[httplib.NOT_FOUND])

    def _zones_ffb62570_8414_4578_a346_526b44e320b7(self, method, url, body,
                                                    headers):
        body = self.fixtures.load('zone_example_com.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _zones_ffb62570_8414_4578_a346_526b44e320b7_records(self, method, url,
                                                            body, headers):
        if method == 'POST':
            body = self.fixtures.load('zone_example_com_record_localhost.json')
        else:
            body = self.fixtures.load('zone_example_com_records.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _zones_ffb62570_8414_4578_a346_526b44e320b7_health_checks(self, method,
                                                                  url, body,
                                                                  headers):
        if method == 'POST':
            body = self.fixtures.load('zone_example_com_health_check.json')
        else:
            body = self.fixtures.load('zone_example_com_health_checks.json')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _zones_1(self, method, url, body, headers):
        return (httplib.NOT_FOUND, body, {},
                httplib.responses[httplib.NOT_FOUND])

    def _zones_1_records(self, method, url, body, headers):
        return (httplib.NOT_FOUND, body, {},
                httplib.responses[httplib.NOT_FOUND])

    def _zones_1_records_1(self, method, url, body, headers):
        return (httplib.NOT_FOUND, body, {},
                httplib.responses[httplib.NOT_FOUND])

    def _zones_ffb62570_8414_4578_a346_526b44e320b7_records_5592f1ff(self,
                                                                     method,
                                                                     url,
                                                                     body,
                                                                     headers):
        body = None
        if method == 'GET':
            body = self.fixtures.load('zone_example_com_record_localhost.json')

        return (httplib.OK, body, {}, httplib.responses[httplib.OK])


if __name__ == '__main__':
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance withv
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from libcloud.utils.py3 import ET
from libcloud.common.dimensiondata import DimensionDataConnection
from libcloud.common.dimensiondata import DimensionDataPool
from libcloud.common.dimensiondata import DimensionDataPoolMember
from libcloud.common.dimensiondata import DimensionDataVirtualListener
from libcloud.common.dimensiondata import DimensionDataVIPNode
from libcloud.common.dimensiondata import DimensionDataDefaultHealthMonitor
from libcloud.common.dimensiondata import DimensionDataPersistenceProfile
from libcloud.common.dimensiondata import \
    DimensionDataVirtualListenerCompatibility
from libcloud.common.dimensiondata import DimensionDataDefaultiRule
from libcloud.common.dimensiondata import API_ENDPOINTS
from libcloud.common.dimensiondata import DEFAULT_REGION
from libcloud.common.dimensiondata import TYPES_URN
from libcloud.utils.misc import reverse_dict
from libcloud.utils.xml import fixxpath, findtext, findall
from libcloud.loadbalancer.types import State
from libcloud.loadbalancer.base import Algorithm, Driver, LoadBalancer
from libcloud.loadbalancer.base import Member
from libcloud.loadbalancer.types import Provider


class DimensionDataLBDriver(Driver):
    """
    DimensionData node driver.
    """

    selected_region = None
    connectionCls = DimensionDataConnection
    name = 'Dimension Data Load Balancer'
    website = 'https://cloud.dimensiondata.com/'
    type = Provider.DIMENSIONDATA
    api_version = 1.0

    network_domain_id = None

    _VALUE_TO_ALGORITHM_MAP = {
        'ROUND_ROBIN': Algorithm.ROUND_ROBIN,
        'LEAST_CONNECTIONS': Algorithm.LEAST_CONNECTIONS,
        'SHORTEST_RESPONSE': Algorithm.SHORTEST_RESPONSE,
        'PERSISTENT_IP': Algorithm.PERSISTENT_IP
    }
    _ALGORITHM_TO_VALUE_MAP = reverse_dict(_VALUE_TO_ALGORITHM_MAP)

    _VALUE_TO_STATE_MAP = {
        'NORMAL': State.RUNNING,
        'PENDING_ADD': State.PENDING,
        'PENDING_CHANGE': State.PENDING,
        'PENDING_DELETE': State.PENDING,
        'FAILED_ADD': State.ERROR,
        'FAILED_CHANGE': State.ERROR,
        'FAILED_DELETE': State.ERROR,
        'REQUIRES_SUPPORT': State.ERROR
    }

    def __init__(self, key, secret=None, secure=True, host=None, port=None,
                 api_version=None, region=DEFAULT_REGION, **kwargs):

        if region not in API_ENDPOINTS and host is None:
            raise ValueError(
                'Invalid region: %s, no host specified' % (region))
        if region is not None:
            self.selected_region = API_ENDPOINTS[region]

        super(DimensionDataLBDriver, self).__init__(key=key, secret=secret,
                                                    secure=secure, host=host,
                                                    port=port,
                                                    api_version=api_version,
                                                    region=region,
                                                    **kwargs)

    def _ex_connection_class_kwargs(self):
        """
            Add the region to the kwargs before the connection is instantiated
        """

        kwargs = super(DimensionDataLBDriver,
                       self)._ex_connection_class_kwargs()
        kwargs['region'] = self.selected_region
        return kwargs

    def create_balancer(self, name, port=None, protocol=None,
                        algorithm=None, members=None,
                        ex_listener_ip_address=None):
        """
        Create a new load balancer instance

        :param name: Name of the new load balancer (required)
        :type  name: ``str``

        :param port: An integer in the range of 1-65535. If not supplied,
                     it will be taken to mean 'Any Port'
        :type  port: ``int``

        :param protocol: Loadbalancer protocol, defaults to http.
        :type  protocol: ``str``

        :param members: list of Members to attach to balancer (optional)
        :type  members: ``list`` of :class:`Member`

        :param algorithm: Load balancing algorithm, defaults to ROUND_ROBIN.
        :type algorithm: :class:`.Algorithm`

        :param ex_listener_ip_address: Must be a valid IPv4 in dot-decimal
                                       notation (x.x.x.x).
        :type ex_listener_ip_address: ``str``

        :rtype: :class:`LoadBalancer`
        """
        network_domain_id = self.network_domain_id
        if protocol is None:
            protocol = 'http'
        if algorithm is None:
            algorithm = Algorithm.ROUND_ROBIN

        # Create a pool first
        pool = self.ex_create_pool(
            network_domain_id=network_domain_id,
            name=name,
            ex_description=None,
            balancer_method=self._ALGORITHM_TO_VALUE_MAP[algorithm])

        # Attach the members to the pool as nodes
        if members is not None:
            for member in members:
                node = self.ex_create_node(
                    network_domain_id=network_domain_id,
                    name=member.ip,
                    ip=member.ip,
                    ex_description=None)
                self.ex_create_pool_member(
                    pool=pool,
                    node=node,
                    port=port)

        # Create the virtual listener (balancer)
        listener = self.ex_create_virtual_listener(
            network_domain_id=network_domain_id,
            name=name,
            ex_description=name,
            port=port,
            pool=pool,
            listener_ip_address=ex_listener_ip_address)

        return LoadBalancer(
            id=listener.id,
            name=listener.name,
            state=State.RUNNING,
            ip=listener.ip,
            port=port,
            driver=self,
            extra={'pool_id': pool.id,
                   'network_domain_id': network_domain_id,
                   'listener_ip_address': ex_listener_ip_address}
        )

    def list_balancers(self, ex_network_domain_id=None):
        """
        List all loadbalancers inside a geography or in given network.

        In Dimension Data terminology these are known as virtual listeners

        :param ex_network_domain_id: UUID of Network Domain
               if not None returns only balancers in the given network
               if None then returns all pools for the organization
        :type  ex_network_domain_id: ``str``

        :rtype: ``list`` of :class:`LoadBalancer`
        """
        params = None
        if ex_network_domain_id is not None:
            params = {"networkDomainId": ex_network_domain_id}

        return self._to_balancers(
            self.connection
            .request_with_orgId_api_2('networkDomainVip/virtualListener',
                                      params=params).object)

    def get_balancer(self, balancer_id):
        """
        Return a :class:`LoadBalancer` object.

        :param balancer_id: id of a load balancer you want to fetch
        :type  balancer_id: ``str``

        :rtype: :class:`LoadBalancer`
        """

        bal = self.connection \
            .request_with_orgId_api_2('networkDomainVip/virtualListener/%s'
                                      % balancer_id).object
        return self._to_balancer(bal)

    def list_protocols(self):
        """
        Return a list of supported protocols.

        Since all protocols are support by Dimension Data, this is a list
        of common protocols.

        :rtype: ``list`` of ``str``
        """
        return ['http', 'https', 'tcp', 'udp', 'ftp', 'smtp']

    def balancer_list_members(self, balancer):
        """
        Return list of members attached to balancer.

        In Dimension Data terminology these are the members of the pools
        within a virtual listener.

        :param balancer: LoadBalancer which should be used
        :type  balancer: :class:`LoadBalancer`

        :rtype: ``list`` of :class:`Member`
        """
        pool_members = self.ex_get_pool_members(balancer.extra['pool_id'])
        members = []
        for pool_member in pool_members:
            members.append(Member(
                id=pool_member.id,
                ip=pool_member.ip,
                port=pool_member.port,
                balancer=balancer,
                extra=None
            ))
        return members

    def balancer_attach_member(self, balancer, member):
        """
        Attach a member to balancer

        :param balancer: LoadBalancer which should be used
        :type  balancer: :class:`LoadBalancer`

        :param member: Member to join to the balancer
        :type member: :class:`Member`

        :return: Member after joining the balancer.
        :rtype: :class:`Member`
        """
        node = self.ex_create_node(
            network_domain_id=balancer.extra['network_domain_id'],
            name='Member.' + member.ip,
            ip=member.ip,
            ex_description=''
        )
        if node is False:
            return False
        pool = self.ex_get_pool(balancer.extra['pool_id'])
        pool_member = self.ex_create_pool_member(
            pool=pool,
            node=node,
            port=member.port)
        member.id = pool_member.id
        return member

    def balancer_detach_member(self, balancer, member):
        """
        Detach member from balancer

        :param balancer: LoadBalancer which should be used
        :type  balancer: :class:`LoadBalancer`

        :param member: Member which should be used
        :type member: :class:`Member`

        :return: ``True`` if member detach was successful, otherwise ``False``.
        :rtype: ``bool``
        """
        create_pool_m = ET.Element('removePoolMember', {'xmlns': TYPES_URN,
                                                        'id': member.id})

        result = self.connection.request_with_orgId_api_2(
            'networkDomainVip/removePoolMember',
            method='POST',
            data=ET.tostring(create_pool_m)).object
        response_code = findtext(result, 'responseCode', TYPES_URN)
        return response_code in ['IN_PROGRESS', 'OK']

    def destroy_balancer(self, balancer):
        """
        Destroy a load balancer (virtual listener)

        :param balancer: LoadBalancer which should be used
        :type  balancer: :class:`LoadBalancer`

        :return: ``True`` if the destroy was successful, otherwise ``False``.
        :rtype: ``bool``
        """
        delete_listener = ET.Element('deleteVirtualListener',
                                     {'xmlns': TYPES_URN,
                                      'id': balancer.id})

        result = self.connection.request_with_orgId_api_2(
            'networkDomainVip/deleteVirtualListener',
            method='POST',
            data=ET.tostring(delete_listener)).object
        response_code = findtext(result, 'responseCode', TYPES_URN)
        return response_code in ['IN_PROGRESS', 'OK']

    def ex_set_current_network_domain(self, network_domain_id):
        """
        Set the network domain (part of the network) of the driver

        :param network_domain_id: ID of the pool (required)
        :type  network_domain_id: ``str``
        """
        self.network_domain_id = network_domain_id

    def ex_get_current_network_domain(self):
        """
        Get the current network domain ID of the driver.

        :return: ID of the network domain
        :rtype: ``str``
        """
        return self.network_domain_id

    def ex_create_pool_member(self, pool, node, port=None):
        """
        Create a new member in an existing pool from an existing node

        :param pool: Instance of ``DimensionDataPool`` (required)
        :type  pool: ``DimensionDataPool``

        :param node: Instance of ``DimensionDataVIPNode`` (required)
        :type  node: ``DimensionDataVIPNode``

        :param port: Port the the service will listen on
        :type  port: ``str``

        :return: The node member, instance of ``DimensionDataPoolMember``
        :rtype: ``DimensionDataPoolMember``
        """
        create_pool_m = ET.Element('addPoolMember', {'xmlns': TYPES_URN})
        ET.SubElement(create_pool_m, "poolId").text = pool.id
        ET.SubElement(create_pool_m, "nodeId").text = node.id
        if port is not None:
            ET.SubElement(create_pool_m, "port").text = str(port)
        ET.SubElement(create_pool_m, "status").text = 'ENABLED'

        response = self.connection.request_with_orgId_api_2(
            'networkDomainVip/addPoolMember',
            method='POST',
            data=ET.tostring(create_pool_m)).object

        member_id = None
        node_name = None
        for info in findall(response, 'info', TYPES_URN):
            if info.get('name') == 'poolMemberId':
                member_id = info.get('value')
            if info.get('name') == 'nodeName':
                node_name = info.get('value')

        return DimensionDataPoolMember(
            id=member_id,
            name=node_name,
            status=State.RUNNING,
            ip=node.ip,
            port=port,
            node_id=node.id
        )

    def ex_create_node(self,
                       network_domain_id,
                       name,
                       ip,
                       ex_description,
                       connection_limit=25000,
                       connection_rate_limit=2000):
        """
        Create a new node

        :param network_domain_id: Network Domain ID (required)
        :type  name: ``str``

        :param name: name of the node (required)
        :type  name: ``str``

        :param ip: IPv4 address of the node (required)
        :type  ip: ``str``

        :param ex_description: Description of the node (required)
        :type  ex_description: ``str``

        :param connection_limit: Maximum number
                of concurrent connections per sec
        :type  connection_limit: ``int``

        :param connection_rate_limit: Maximum number of concurrent sessions
        :type  connection_rate_limit: ``int``

        :return: Instance of ``DimensionDataVIPNode``
        :rtype: ``DimensionDataVIPNode``
        """
        create_node_elm = ET.Element('createNode', {'xmlns': TYPES_URN})
        ET.SubElement(create_node_elm, "networkDomainId") \
            .text = network_domain_id
        ET.SubElement(create_node_elm, "name").text = name
        ET.SubElement(create_node_elm, "description").text \
            = str(ex_description)
        ET.SubElement(create_node_elm, "ipv4Address").text = ip
        ET.SubElement(create_node_elm, "status").text = 'ENABLED'
        ET.SubElement(create_node_elm, "connectionLimit") \
            .text = str(connection_limit)
        ET.SubElement(create_node_elm, "connectionRateLimit") \
            .text = str(connection_rate_limit)

        response = self.connection.request_with_orgId_api_2(
            action='networkDomainVip/createNode',
            method='POST',
            data=ET.tostring(create_node_elm)).object

        node_id = None
        node_name = None
        for info in findall(response, 'info', TYPES_URN):
            if info.get('name') == 'nodeId':
                node_id = info.get('value')
            if info.get('name') == 'name':
                node_name = info.get('value')
        return DimensionDataVIPNode(
            id=node_id,
            name=node_name,
            status=State.RUNNING,
            ip=ip
        )

    def ex_update_node(self, node):
        """
        Update the properties of a node

        :param pool: The instance of ``DimensionDataNode`` to update
        :type  pool: ``DimensionDataNode``

        :return: The instance of ``DimensionDataNode``
        :rtype: ``DimensionDataNode``
        """
        create_node_elm = ET.Element('editNode', {'xmlns': TYPES_URN})
        ET.SubElement(create_node_elm, "connectionLimit") \
            .text = str(node.connection_limit)
        ET.SubElement(create_node_elm, "connectionRateLimit") \
            .text = str(node.connection_rate_limit)

        self.connection.request_with_orgId_api_2(
            action='networkDomainVip/createNode',
            method='POST',
            data=ET.tostring(create_node_elm)).object
        return node

    def ex_set_node_state(self, node, enabled):
        """
        Change the state of a node (enable/disable)

        :param pool: The instance of ``DimensionDataNode`` to update
        :type  pool: ``DimensionDataNode``

        :param enabled: The target state of the node
        :type  enabled: ``bool``

        :return: The instance of ``DimensionDataNode``
        :rtype: ``DimensionDataNode``
        """
        create_node_elm = ET.Element('editNode', {'xmlns': TYPES_URN})
        ET.SubElement(create_node_elm, "status") \
            .text = "ENABLED" if enabled is True else "DISABLED"

        self.connection.request_with_orgId_api_2(
            action='networkDomainVip/editNode',
            method='POST',
            data=ET.tostring(create_node_elm)).object
        return node

    def ex_create_pool(self,
                       network_domain_id,
                       name,
                       balancer_method,
                       ex_description,
                       health_monitors=None,
                       service_down_action='NONE',
                       slow_ramp_time=30):
        """
        Create a new pool

        :param network_domain_id: Network Domain ID (required)
        :type  name: ``str``

        :param name: name of the node (required)
        :type  name: ``str``

        :param balancer_method: The load balancer algorithm (required)
        :type  balancer_method: ``str``

        :param ex_description: Description of the node (required)
        :type  ex_description: ``str``

        :param health_monitors: A list of health monitors to use for the pool.
        :type  health_monitors: ``list`` of
            :class:`DimensionDataDefaultHealthMonitor`

        :param service_down_action: What to do when node
                                    is unavailable NONE, DROP or RESELECT
        :type  service_down_action: ``str``

        :param slow_ramp_time: Number of seconds to stagger ramp up of nodes
        :type  slow_ramp_time: ``int``

        :return: Instance of ``DimensionDataPool``
        :rtype: ``DimensionDataPool``
        """
        # Names cannot contain spaces.
        name.replace(' ', '_')
        create_node_elm = ET.Element('createPool', {'xmlns': TYPES_URN})
        ET.SubElement(create_node_elm, "networkDomainId") \
            .text = network_domain_id
        ET.SubElement(create_node_elm, "name").text = name
        ET.SubElement(create_node_elm, "description").text \
            = str(ex_description)
        ET.SubElement(create_node_elm, "loadBalanceMethod") \
            .text = str(balancer_method)

        if health_monitors is not None:
            for monitor in health_monitors:
                ET.SubElement(create_node_elm, "healthMonitorId") \
                    .text = str(monitor.id)

        ET.SubElement(create_node_elm, "serviceDownAction") \
            .text = service_down_action
        ET.SubElement(create_node_elm, "slowRampTime").text \
            = str(slow_ramp_time)

        response = self.connection.request_with_orgId_api_2(
            action='networkDomainVip/createPool',
            method='POST',
            data=ET.tostring(create_node_elm)).object

        pool_id = None
        for info in findall(response, 'info', TYPES_URN):
            if info.get('name') == 'poolId':
                pool_id = info.get('value')

        return DimensionDataPool(
            id=pool_id,
            name=name,
            description=ex_description,
            status=State.RUNNING,
            load_balance_method=str(balancer_method),
            health_monitor_id=None,
            service_down_action=service_down_action,
            slow_ramp_time=str(slow_ramp_time)
        )

    def ex_create_virtual_listener(self,
                                   network_domain_id,
                                   name,
                                   ex_description,
                                   port=None,
                                   pool=None,
                                   listener_ip_address=None,
                                   persistence_profile=None,
                                   fallback_persistence_profile=None,
                                   irule=None,
                                   protocol='TCP',
                                   connection_limit=25000,
                                   connection_rate_limit=2000,
                                   source_port_preservation='PRESERVE'):
        """
        Create a new virtual listener (load balancer)

        :param network_domain_id: Network Domain ID (required)
        :type  name: ``str``

        :param name: name of the listener (required)
        :type  name: ``str``

        :param ex_description: Description of the node (required)
        :type  ex_description: ``str``

        :param port: An integer in the range of 1-65535. If not supplied,
                     it will be taken to mean 'Any Port'
        :type  port: ``int``

        :param pool: The pool to use for the listener
        :type  pool: :class:`DimensionDataPool`

        :param listener_ip_address: The IPv4 Address of the virtual listener
        :type  listener_ip_address: ``str``

        :param persistence_profile: Persistence profile
        :type  persistence_profile: :class:`DimensionDataPersistenceProfile`

        :param fallback_persistence_profile: Fallback persistence profile
        :type  fallback_persistence_profile:
            :class:`DimensionDataPersistenceProfile`

        :param irule: The iRule to apply
        :type  irule: :class:`DimensionDataDefaultiRule`

        :param protocol: For STANDARD type, ANY, TCP or UDP
                         for PERFORMANCE_LAYER_4 choice of ANY, TCP, UDP, HTTP
        :type  protcol: ``str``

        :param connection_limit: Maximum number
                                of concurrent connections per sec
        :type  connection_limit: ``int``

        :param connection_rate_limit: Maximum number of concurrent sessions
        :type  connection_rate_limit: ``int``

        :param source_port_preservation: Choice of PRESERVE,
                                        PRESERVE_STRICT or CHANGE
        :type  source_port_preservation: ``str``

        :return: Instance of the listener
        :rtype: ``DimensionDataVirtualListener``
        """
        if (port == 80) or (port == 443):
            listener_type = 'PERFORMANCE_LAYER_4'
        else:
            listener_type = 'STANDARD'

        create_node_elm = ET.Element('createVirtualListener',
                                     {'xmlns': TYPES_URN})
        ET.SubElement(create_node_elm, "networkDomainId") \
            .text = network_domain_id
        ET.SubElement(create_node_elm, "name").text = name
        ET.SubElement(create_node_elm, "description").text = \
            str(ex_description)
        ET.SubElement(create_node_elm, "type").text = listener_type
        ET.SubElement(create_node_elm, "protocol") \
            .text = protocol
        if listener_ip_address is not None:
            ET.SubElement(create_node_elm, "listenerIpAddress").text = \
                str(listener_ip_address)
        if port is not None:
            ET.SubElement(create_node_elm, "port").text = str(port)
        ET.SubElement(create_node_elm, "enabled").text = 'true'
        ET.SubElement(create_node_elm, "connectionLimit") \
            .text = str(connection_limit)
        ET.SubElement(create_node_elm, "connectionRateLimit") \
            .text = str(connection_rate_limit)
        ET.SubElement(create_node_elm, "sourcePortPreservation") \
            .text = source_port_preservation
        if pool is not None:
            ET.SubElement(create_node_elm, "poolId") \
                .text = pool.id
        if persistence_profile is not None:
            ET.SubElement(create_node_elm, "persistenceProfileId") \
                .text = persistence_profile.id
        if fallback_persistence_profile is not None:
            ET.SubElement(create_node_elm, "fallbackPersistenceProfileId") \
                .text = fallback_persistence_profile.id
        if irule is not None:
            ET.SubElement(create_node_elm, "iruleId") \
                .text = irule.id

        response = self.connection.request_with_orgId_api_2(
            action='networkDomainVip/createVirtualListener',
            method='POST',
            data=ET.tostring(create_node_elm)).object

        virtual_listener_id = None
        virtual_listener_ip = None
        for info in findall(response, 'info', TYPES_URN):
            if info.get('name') == 'virtualListenerId':
                virtual_listener_id = info.get('value')
            if info.get('name') == 'listenerIpAddress':
                virtual_listener_ip = info.get('value')

        return DimensionDataVirtualListener(
            id=virtual_listener_id,
            name=name,
            ip=virtual_listener_ip,
            status=State.RUNNING
        )

    def ex_get_pools(self, ex_network_domain_id=None):
        """
        Get all of the pools inside the current geography or
        in given network.

        :param ex_network_domain_id: UUID of Network Domain
               if not None returns only balancers in the given network
               if None then returns all pools for the organization
        :type  ex_network_domain_id: ``str``

        :return: Returns a ``list`` of type ``DimensionDataPool``
        :rtype: ``list`` of ``DimensionDataPool``
        """
        params = None

        if ex_network_domain_id is not None:
            params = {"networkDomainId": ex_network_domain_id}

        pools = self.connection \
            .request_with_orgId_api_2('networkDomainVip/pool',
                                      params=params).object

        return self._to_pools(pools)

    def ex_get_pool(self, pool_id):
        """
        Get a specific pool inside the current geography

        :param pool_id: The identifier of the pool
        :type  pool_id: ``str``

        :return: Returns an instance of ``DimensionDataPool``
        :rtype: ``DimensionDataPool``
        """
        pool = self.connection \
            .request_with_orgId_api_2('networkDomainVip/pool/%s'
                                      % pool_id).object
        return self._to_pool(pool)

    def ex_update_pool(self, pool):
        """
        Update the properties of an existing pool
        only method, serviceDownAction and slowRampTime are updated

        :param pool: The instance of ``DimensionDataPool`` to update
        :type  pool: ``DimensionDataPool``

        :return: ``True`` for success, ``False`` for failure
        :rtype: ``bool``
        """
        create_node_elm = ET.Element('editPool', {'xmlns': TYPES_URN})

        ET.SubElement(create_node_elm, "loadBalanceMethod") \
            .text = str(pool.load_balance_method)
        ET.SubElement(create_node_elm, "serviceDownAction") \
            .text = pool.service_down_action
        ET.SubElement(create_node_elm, "slowRampTime").text \
            = str(pool.slow_ramp_time)

        response = self.connection.request_with_orgId_api_2(
            action='networkDomainVip/editPool',
            method='POST',
            data=ET.tostring(create_node_elm)).object
        response_code = findtext(response, 'responseCode', TYPES_URN)
        return response_code in ['IN_PROGRESS', 'OK']

    def ex_destroy_pool(self, pool):
        """
        Destroy an existing pool

        :param pool: The instance of ``DimensionDataPool`` to destroy
        :type  pool: ``DimensionDataPool``

        :return: ``True`` for success, ``False`` for failure
        :rtype: ``bool``
        """
        destroy_request = ET.Element('deletePool',
                                     {'xmlns': TYPES_URN,
                                      'id': pool.id})

        result = self.connection.request_with_orgId_api_2(
            action='networkDomainVip/deletePool',
            method='POST',
            data=ET.tostring(destroy_request)).object
        response_code = findtext(result, 'responseCode', TYPES_URN)
        return response_code in ['IN_PROGRESS', 'OK']

    def ex_get_pool_members(self, pool_id):
        """
        Get the members of a pool

        :param pool: The instance of a pool
        :type  pool: ``DimensionDataPool``

        :return: Returns an ``list`` of ``DimensionDataPoolMember``
        :rtype: ``list`` of ``DimensionDataPoolMember``
        """
        members = self.connection \
            .request_with_orgId_api_2('networkDomainVip/poolMember?poolId=%s'
                                      % pool_id).object
        return self._to_members(members)

    def ex_get_pool_member(self, pool_member_id):
        """
        Get a specific member of a pool

        :param pool: The id of a pool member
        :type  pool: ``str``

        :return: Returns an instance of ``DimensionDataPoolMember``
        :rtype: ``DimensionDataPoolMember``
        """
        member = self.connection \
            .request_with_orgId_api_2('networkDomainVip/poolMember/%s'
                                      % pool_member_id).object
        return self._to_member(member)

    def ex_set_pool_member_state(self, member, enabled=True):
        request = ET.Element('editPoolMember',
                             {'xmlns': TYPES_URN,
                              'id': member.id})
        state = "ENABLED" if enabled is True else "DISABLED"
        ET.SubElement(request, 'status').text = state

        result = self.connection.request_with_orgId_api_2(
            action='networkDomainVip/editPoolMember',
            method='POST',
            data=ET.tostring(request)).object

        response_code = findtext(result, 'responseCode', TYPES_URN)
        return response_code in ['IN_PROGRESS', 'OK']

    def ex_destroy_pool_member(self, member, destroy_node=False):
        """
        Destroy a specific member of a pool

        :param pool: The instance of a pool member
        :type  pool: ``DimensionDataPoolMember``

        :param destroy_node: Also destroy the associated node
        :type  destroy_node: ``bool``

        :return: ``True`` for success, ``False`` for failure
        :rtype: ``bool``
        """
        # remove the pool member
        destroy_request = ET.Element('removePoolMember',
                                     {'xmlns': TYPES_URN,
                                      'id': member.id})

        result = self.connection.request_with_orgId_api_2(
            action='networkDomainVip/removePoolMember',
            method='POST',
            data=ET.tostring(destroy_request)).object

        if member.node_id is not None and destroy_node is True:
            return self.ex_destroy_node(member.node_id)
        else:
            response_code = findtext(result, 'responseCode', TYPES_URN)
            return response_code in ['IN_PROGRESS', 'OK']

    def ex_get_nodes(self, ex_network_domain_id=None):
        """
        Get the nodes within this geography or in given network.

        :param ex_network_domain_id: UUID of Network Domain
               if not None returns only balancers in the given network
               if None then returns all pools for the organization
        :type  ex_network_domain_id: ``str``

        :return: Returns an ``list`` of ``DimensionDataVIPNode``
        :rtype: ``list`` of ``DimensionDataVIPNode``
        """
        params = None
        if ex_network_domain_id is not None:
            params = {"networkDomainId": ex_network_domain_id}

        nodes = self.connection \
            .request_with_orgId_api_2('networkDomainVip/node',
                                      params=params).object
        return self._to_nodes(nodes)

    def ex_get_node(self, node_id):
        """
        Get the node specified by node_id

        :return: Returns an instance of ``DimensionDataVIPNode``
        :rtype: Instance of ``DimensionDataVIPNode``
        """
        nodes = self.connection \
            .request_with_orgId_api_2('networkDomainVip/node/%s'
                                      % node_id).object
        return self._to_node(nodes)

    def ex_destroy_node(self, node_id):
        """
        Destroy a specific node

        :param node_id: The ID of of a ``DimensionDataVIPNode``
        :type  node_id: ``str``

        :return: ``True`` for success, ``False`` for failure
        :rtype: ``bool``
        """
        # Destroy the node
        destroy_request = ET.Element('deleteNode',
                                     {'xmlns': TYPES_URN,
                                      'id': node_id})

        result = self.connection.request_with_orgId_api_2(
            action='networkDomainVip/deleteNode',
            method='POST',
            data=ET.tostring(destroy_request)).object
        response_code = findtext(result, 'responseCode', TYPES_URN)
        return response_code in ['IN_PROGRESS', 'OK']

    def ex_wait_for_state(self, state, func, poll_interval=2,
                          timeout=60, *args, **kwargs):
        """
        Wait for the function which returns a instance
        with field status to match

        Keep polling func until one of the desired states is matched

        :param state: Either the desired state (`str`) or a `list` of states
        :type  state: ``str`` or ``list``

        :param  func: The function to call, e.g. ex_get_vlan
        :type   func: ``function``

        :param  poll_interval: The number of seconds to wait between checks
        :type   poll_interval: `int`

        :param  timeout: The total number of seconds to wait to reach a state
        :type   timeout: `int`

        :param  args: The arguments for func
        :type   args: Positional arguments

        :param  kwargs: The arguments for func
        :type   kwargs: Keyword arguments
        """
        return self.connection.wait_for_state(state, func, poll_interval,
                                              timeout, *args, **kwargs)

    def ex_get_default_health_monitors(self, network_domain_id):
        """
        Get the default health monitors available for a network domain

        :param network_domain_id: The ID of of a ``DimensionDataNetworkDomain``
        :type  network_domain_id: ``str``

        :rtype: `list` of :class:`DimensionDataDefaultHealthMonitor`
        """
        result = self.connection.request_with_orgId_api_2(
            action='networkDomainVip/defaultHealthMonitor',
            params={'networkDomainId': network_domain_id},
            method='GET').object
        return self._to_health_monitors(result)

    def ex_get_default_persistence_profiles(self, network_domain_id):
        """
        Get the default persistence profiles available for a network domain

        :param network_domain_id: The ID of of a ``DimensionDataNetworkDomain``
        :type  network_domain_id: ``str``

        :rtype: `list` of :class:`DimensionDataPersistenceProfile`
        """
        result = self.connection.request_with_orgId_api_2(
            action='networkDomainVip/defaultPersistenceProfile',
            params={'networkDomainId': network_domain_id},
            method='GET').object
        return self._to_persistence_profiles(result)

    def ex_get_default_irules(self, network_domain_id):
        """
        Get the default iRules available for a network domain

        :param network_domain_id: The ID of of a ``DimensionDataNetworkDomain``
        :type  network_domain_id: ``str``

        :rtype: `list` of :class:`DimensionDataDefaultiRule`
        """
        result = self.connection.request_with_orgId_api_2(
            action='networkDomainVip/defaultIrule',
            params={'networkDomainId': network_domain_id},
            method='GET').object
        return self._to_irules(result)

    def _to_irules(self, object):
        irules = []
        matches = object.findall(
            fixxpath('defaultIrule', TYPES_URN))
        for element in matches:
            irules.append(self._to_irule(element))
        return irules

    def _to_irule(self, element):
        compatible = []
        matches = element.findall(
            fixxpath('virtualListenerCompatibility', TYPES_URN))
        for match_element in matches:
            compatible.append(
                DimensionDataVirtualListenerCompatibility(
                    type=match_element.get('type'),
                    protocol=match_element.get('protocol', None)))
        irule_element = element.find(fixxpath('irule', TYPES_URN))
        return DimensionDataDefaultiRule(
            id=irule_element.get('id'),
            name=irule_element.get('name'),
            compatible_listeners=compatible
        )

    def _to_persistence_profiles(self, object):
        profiles = []
        matches = object.findall(
            fixxpath('defaultPersistenceProfile', TYPES_URN))
        for element in matches:
            profiles.append(self._to_persistence_profile(element))
        return profiles

    def _to_persistence_profile(self, element):
        compatible = []
        matches = element.findall(
            fixxpath('virtualListenerCompatibility', TYPES_URN))
        for match_element in matches:
            compatible.append(
                DimensionDataVirtualListenerCompatibility(
                    type=match_element.get('type'),
                    protocol=match_element.get('protocol', None)))

        return DimensionDataPersistenceProfile(
            id=element.get('id'),
            fallback_compatible=bool(
                element.get('fallbackCompatible') == "true"),
            name=findtext(element, 'name', TYPES_URN),
            compatible_listeners=compatible
        )

    def _to_health_monitors(self, object):
        monitors = []
        matches = object.findall(fixxpath('defaultHealthMonitor', TYPES_URN))
        for element in matches:
            monitors.append(self._to_health_monitor(element))
        return monitors

    def _to_health_monitor(self, element):
        return DimensionDataDefaultHealthMonitor(
            id=element.get('id'),
            name=findtext(element, 'name', TYPES_URN),
            node_compatible=bool(
                findtext(element, 'nodeCompatible', TYPES_URN) == "true"),
            pool_compatible=bool(
                findtext(element, 'poolCompatible', TYPES_URN) == "true"),
        )

    def _to_nodes(self, object):
        nodes = []
        for element in object.findall(fixxpath("node", TYPES_URN)):
            nodes.append(self._to_node(element))

        return nodes

    def _to_node(self, element):
        ipaddress = findtext(element, 'ipv4Address', TYPES_URN)
        if ipaddress is None:
            ipaddress = findtext(element, 'ipv6Address', TYPES_URN)

        name = findtext(element, 'name', TYPES_URN)

        node = DimensionDataVIPNode(
            id=element.get('id'),
            name=name,
            status=self._VALUE_TO_STATE_MAP.get(
                findtext(element, 'state', TYPES_URN),
                State.UNKNOWN),
            connection_rate_limit=findtext(element,
                                           'connectionRateLimit', TYPES_URN),
            connection_limit=findtext(element, 'connectionLimit', TYPES_URN),
            ip=ipaddress)

        return node

    def _to_balancers(self, object):
        loadbalancers = []
        for element in object.findall(fixxpath("virtualListener", TYPES_URN)):
            loadbalancers.append(self._to_balancer(element))

        return loadbalancers

    def _to_balancer(self, element):
        ipaddress = findtext(element, 'listenerIpAddress', TYPES_URN)
        name = findtext(element, 'name', TYPES_URN)
        port = findtext(element, 'port', TYPES_URN)
        extra = {}

        pool_element = element.find(fixxpath(
            'pool',
            TYPES_URN))
        if pool_element is None:
            extra['pool_id'] = None

        else:
            extra['pool_id'] = pool_element.get('id')

        extra['network_domain_id'] = findtext(element, 'networkDomainId',
                                              TYPES_URN)

        balancer = LoadBalancer(
            id=element.get('id'),
            name=name,
            state=self._VALUE_TO_STATE_MAP.get(
                findtext(element, 'state', TYPES_URN),
                State.UNKNOWN),
            ip=ipaddress,
            port=port,
            driver=self.connection.driver,
            extra=extra
        )

        return balancer

    def _to_members(self, object):
        members = []
        for element in object.findall(fixxpath("poolMember", TYPES_URN)):
            members.append(self._to_member(element))

        return members

    def _to_member(self, element):
        port = findtext(element, 'port', TYPES_URN)
        if port is not None:
            port = int(port)
        pool_member = DimensionDataPoolMember(
            id=element.get('id'),
            name=element.find(fixxpath(
                'node',
                TYPES_URN)).get('name'),
            status=findtext(element, 'state', TYPES_URN),
            node_id=element.find(fixxpath(
                'node',
                TYPES_URN)).get('id'),
            ip=element.find(fixxpath(
                'node',
                TYPES_URN)).get('ipAddress'),
            port=port
        )
        return pool_member

    def _to_pools(self, object):
        pools = []
        for element in object.findall(fixxpath("pool", TYPES_URN)):
            pools.append(self._to_pool(element))

        return pools

    def _to_pool(self, element):
        pool = DimensionDataPool(
            id=element.get('id'),
            name=findtext(element, 'name', TYPES_URN),
            status=findtext(element, 'state', TYPES_URN),
            description=findtext(element, 'description', TYPES_URN),
            load_balance_method=findtext(element, 'loadBalanceMethod',
                                         TYPES_URN),
            health_monitor_id=findtext(element, 'healthMonitorId', TYPES_URN),
            service_down_action=findtext(element, 'serviceDownAction',
                                         TYPES_URN),
            slow_ramp_time=findtext(element, 'slowRampTime', TYPES_URN),
        )
        return pool
